"""Provides custom callbacks for the Hugging Face Trainer class."""

from transformers import TrainerCallback
from pathlib import Path
import pandas as pd
from copy import deepcopy
import os, shutil


class MetricsLoggerCallback(TrainerCallback):
    def _load_current_file(self, output_dir):
        output_dir = Path(output_dir)
        metrics_file = output_dir / "metrics.csv"

        return (
            pd.read_csv(metrics_file)
            if metrics_file.exists()
            else pd.DataFrame()
        )

    def on_log(self, args, state, control, logs=None, **kwargs):
        """
        Triggered during training logging (e.g., for training loss).
        """
        if "loss" in logs:  # Ensure training loss is available
            metrics_df = self._load_current_file(args.output_dir)

            # Append the epoch and the training loss.
            loss = logs["loss"]
            epoch = state.epoch

            metrics_df = pd.concat(
                [
                    metrics_df,
                    pd.DataFrame([{"epoch": epoch, "train_loss": loss}]),
                ],
                ignore_index=True,
            )

            # Save it back to file.
            metrics_df.to_csv(
                Path(args.output_dir) / "metrics.csv", index=False
            )

    def on_evaluate(self, args, state, control, metrics, **kwargs):
        """
        This is triggered at the end of evaluation (including at the end of each epoch).
        """
        # Get the current metrics file.
        metrics_df = self._load_current_file(args.output_dir)

        # Add metrics to the last row of the current metrics file.
        for k, v in metrics.items():
            metrics_df.at[metrics_df.index[-1], k] = v

        # Save the updated metrics file.
        metrics_df.to_csv(Path(args.output_dir) / "metrics.csv", index=False)


class BestModelCallback(TrainerCallback):
    """Callback for tracking and saving the best model checkpoint.

    This class was developed with the help of ChatGPT."""

    def __init__(self, save_dir: str):
        """
        Initialize the callback.

        Args:
            save_dir (str): Directory where model is saving checkpoints.
                The best checkpoint will be saved in a subdirectory
                called "best-checkpoint".

        """
        self.best_loss = float("inf")
        self.best_model_state = None
        self.save_dir = save_dir

    def on_evaluate(self, args, state, control, metrics, model, **kwargs):
        """Called after evaluation. Track the best model state."""
        # Get the mean_dice metric
        current_metric = metrics.get("eval_loss")

        if current_metric is not None and current_metric < self.best_loss:
            self.best_loss = current_metric
            print("New best loss.")

            # Store the model state.
            self.best_model_state = deepcopy(model.state_dict())

    def on_train_end(self, args, state, control, model, **kwargs):
        """Logic:
        * Create a best-checkpoint directory
        * Save the best model state into the best-checkpoint directory
        * Copy all the other model files (from the normal checkpoint
          directory) into the best-checkpoint directory

        """
        if self.best_model_state is not None:
            # Create best-checkpoint directory
            best_checkpoint_dir = os.path.join(
                self.save_dir, "best-checkpoint"
            )
            os.makedirs(best_checkpoint_dir, exist_ok=True)

            # Load the best model state
            model.load_state_dict(self.best_model_state)

            # Save the model (this creates model.safetensors or pytorch_model.bin)
            model.save_pretrained(best_checkpoint_dir)

            # Copy other checkpoint files from the latest checkpoint
            # Find the latest checkpoint directory (checkpoint-{step_number})
            save_path = Path(self.save_dir)
            checkpoint_dirs = [
                d
                for d in save_path.iterdir()
                if d.is_dir() and d.name.startswith("checkpoint-")
            ]
            if checkpoint_dirs:
                # Sort by step number to get the latest
                latest_checkpoint = max(
                    checkpoint_dirs, key=lambda x: int(x.name.split("-")[1])
                )

                # Copy all files from the latest checkpoint, except the model files we just created
                for file_path in latest_checkpoint.iterdir():
                    if file_path.is_file():
                        # Skip model files since we already saved the best model state
                        if not file_path.name.startswith(
                            ("model.", "pytorch_model.")
                        ):
                            dst = Path(best_checkpoint_dir) / file_path.name
                            shutil.copy2(file_path, dst)



import large_image_source_openslide, ultralytics
import geopandas as gpd
import pandas as pd
import cv2 as cv
import numpy as np
from shapely import Polygon
from time import perf_counter

from ...utils import non_max_suppression, return_mag_and_resolution
from ...gpd_utils import remove_contained_boxes


def yolo_inference_no_iterator(
    model: str | ultralytics.YOLO,
    wsi_fp: str,
    mag: float | None = 20,
    mm_px: float | None = None,
    tile_size: int = 640,
    overlap: int = 160,
    batch_size: int = 64,
    agnostic_nms: bool = True,
    conf_thr: float = 0.25,
    iou: float = 0.7,
    tile_area_threshold: float = 0.25,
    pad_rgb: tuple[int, int, int] = (114, 114, 114),
    device: str | None = None,
):
    """Perform YOLO inference on a whole slide image. This version differs
    from yolo_inference in that it does not use the eager iterator.

    Args:
        model (str | ultralytics.YOLO): YOLO model or path to the model
            weights.
        wsi_fp (str): File path to the whole slide image.
        mag (float | None, optional): Desired magnification for tiling.
            Note magnification meaning can vary among scanners. We
            convert this to a corresponding resolution defined by a
            standard magnification to resolution ratio (Emory scanner).
            Default is None. Defined by mm_px, or if that is None than
            the scan resolution along the x axis is used.
        mm_px (float | None, optional): Micrometers per pixel for the
            desired resolution. If not provided, will be inferred from
            the mag parameter by standardizing to the Emory scanner. If
            both are not provided, the scan resolution in the x
            direction is used. Default is None.
        tile_size (int, optional): Image is tiled into smaller images,
            tiles, of this size at the desired resolution.
        overlap (int, optional): The number of pixels to overlap between
            tiles. Default is 160.
        batch_size (int, optional): The number of tiles to process in a
            single batch. Used both for the eager iterator and the YOLO
            training. Default is 64.
        agnostic_nms (bool, optional): Whether to use agnostic NMS.
            Default is True.
        conf_thr (float, optional): The confidence threshold for the
            NMS. Default is 0.25.
        iou (float, optional): The IoU threshold for the NMS. Default is
            0.7.
        tile_area_threshold (float, optional): The threshold for the tile
            area in region. Default is 0.25.
        pad_rgb (tuple[int, int, int], optional): The RGB values to pad
            the tiles with. Default is (114, 114, 114).
        device (str | None, optional): The device to use for inference.
            Default is None, will use "cuda" if available, otherwise
            "cpu". You can specify values like "cuda:0", "0", or
            multiple devices separated by commas.

    Returns:
        dict: A dictionary containing the inference results.
            - gdf (geopandas.GeoDataFrame): The inference results as a
              GeoDataFrame.
            - mag (float): The magnification used for inference.
            - mm_px (float): The micrometers per pixel used for inference.
            - time (dict): A dictionary containing the time taken for
              inference.
                - total (float): The total time taken for inference.

    Raises:
        ValueError: If both mag and mm_px are provided.

    """
    print("This function has not been reviewed yet, use with caution.")
    start_time = perf_counter()

    if isinstance(model, str):
        model = ultralytics.YOLO(model)

    ts = large_image_source_openslide.open(wsi_fp)
    ts_metadata = ts.getMetadata()

    mm_x = ts_metadata["mm_x"]
    mm_y = ts_metadata["mm_y"]

    # Get the desired resolution.
    if mag is not None and mm_px is not None:
        raise ValueError("Only one of mag or mm_px can be provided.")
    if mag is None and mm_px is None:
        # Use the scan resolution, we use the x resolution.
        mag, mm_px = return_mag_and_resolution(mm_px=mm_x)
    else:
        mag, mm_px = return_mag_and_resolution(mag=mag, mm_px=mm_px)

    # Calculate the x and y size of the tile at scan resolution.
    # desired resolution x sf_* -> scan resolution
    sf_x = mm_px / mm_x
    sf_y = mm_px / mm_y

    scan_tile_x = int(tile_size * sf_x)
    scan_tile_y = int(tile_size * sf_y)

    stride = tile_size - overlap
    scan_stride_x = int(stride * sf_x)
    scan_stride_y = int(stride * sf_y)

    # Calculate the x, y coordinates
    xys = []

    right = ts_metadata["sizeX"]
    bottom = ts_metadata["sizeY"]

    for x in range(0, right, scan_stride_x):
        for y in range(0, bottom, scan_stride_y):
            xys.append((x, y))

    # Process in batches.
    batch_indices = np.arange(0, len(xys), batch_size)

    # Iterate through the WSI.
    boxes = []

    # Calculate the tile area to be above the threshold.
    tile_area = tile_size * tile_size
    minimum_tile_area = tile_area * tile_area_threshold

    for batch_idx in batch_indices:
        xy_index = batch_idx * batch_size

        # Get the tiles.
        batch_xys = xys[xy_index : xy_index + batch_size]

        tiles = []

        tile_x_coords = []
        tile_y_coords = []

        for xy in batch_xys:
            x1, y1 = xy

            x2 = x1 + scan_tile_x

            if x2 > right:
                x2 = right

            y2 = y1 + scan_tile_y

            if y2 > bottom:
                y2 = bottom

            # Get the tiles.
            tile = ts.getRegion(
                region={
                    "left": x1,
                    "top": y1,
                    "right": x2,
                    "bottom": y2,
                },
                format="numpy",
                scale={"mm_x": mm_px, "mm_y": mm_px},
            )[0][:, :, :3].copy()

            # Pad the tile if it is not the right shape.
            tile_h, tile_w, _ = tile.shape

            tile_area = tile_h * tile_w

            if tile_area < minimum_tile_area:
                continue

            tile_x_coords.append(x1)
            tile_y_coords.append(y1)

            if tile_h != tile_size or tile_w != tile_size:
                tile = cv.copyMakeBorder(
                    tile,
                    0,
                    tile_size - tile_h,
                    0,
                    tile_size - tile_w,
                    cv.BORDER_CONSTANT,
                    value=pad_rgb,
                )
            # Convert to BGR.
            tile = cv.cvtColor(tile, cv.COLOR_RGB2BGR)

            tiles.append(tile)

        if len(tiles) == 0:
            continue

        results = model(
            tiles,
            imgsz=tile_size,
            batch=batch_size,
            agnostic_nms=agnostic_nms,
            verbose=False,
            conf=conf_thr,
            iou=iou,
            device=device,
        )

        # Loop through tile variables.
        for result, x, y in zip(results, tile_x_coords, tile_y_coords):
            xyxys = result.boxes.xyxy
            cls_list = result.boxes.cls
            conf_list = result.boxes.conf

            for xyxy, cls, conf in zip(xyxys, cls_list, conf_list):
                # cls and conf are tensorts in device, convert to to int and float
                cls = int(cls)
                conf = float(conf)

                x1, y1, x2, y2 = xyxy

                # (1) scale to scan resolution, (2) shift to tile location.
                x1 = int(x1 * sf_x) + x
                y1 = int(y1 * sf_y) + y
                x2 = int(x2 * sf_x) + x
                y2 = int(y2 * sf_y) + y

                # Create the polygon.
                geom = Polygon([(x1, y1), (x2, y1), (x2, y2), (x1, y2)])
                boxes.append([cls, x1, y1, x2, y2, conf, geom])

    gdf = gpd.GeoDataFrame(
        boxes, columns=["label", "x1", "y1", "x2", "y2", "conf", "geometry"]
    )
    gdf["box_area"] = gdf.geometry.area

    # Clean up predictions further.
    gdf = non_max_suppression(gdf, conf_thr)
    gdf = remove_contained_boxes(gdf, iou).reset_index(drop=True)

    return gdf, mag, mm_px


def yolo_inference(
    model: str | ultralytics.YOLO,
    wsi_fp: str,
    mag: float | None = 20,
    mm_px: float | None = None,
    tile_size: int = 640,
    overlap: int = 160,
    batch_size: int = 64,
    prefetch: int = 2,
    workers: int = 8,
    chunk_mult: int = 2,
    agnostic_nms: bool = True,
    conf_thr: float = 0.25,
    iou: float = 0.7,
    device: str | None = None,
) -> dict:
    """Perform YOLO inference on a whole slide image.

    Args:
        model (str | ultralytics.YOLO): YOLO model or path to the model
            weights.
        wsi_fp (str): File path to the whole slide image.
        mag (float | None, optional): Desired magnification for tiling.
            Note magnification meaning can vary among scanners. We
            convert this to a corresponding resolution defined by a
            standard magnification to resolution ratio (Emory scanner).
            Default is None. Defined by mm_px, or if that is None than
            the scan resolution along the x axis is used.
        mm_px (float | None, optional): Micrometers per pixel for the
            desired resolution. If not provided, will be inferred from
            the mag parameter by standardizing to the Emory scanner. If
            both are not provided, the scan resolution in the x
            direction is used. Default is None.
        tile_size (int, optional): Image is tiled into smaller images,
            tiles, of this size at the desired resolution.
        overlap (int, optional): The number of pixels to overlap between
            tiles. Default is 160.
        batch_size (int, optional): The number of tiles to process in a
            single batch. Used both for the eager iterator and the YOLO
            training. Default is 64.
        prefetch (int, optional): The number of batches to prefetch.
            Used for the eager iterator. Default is 2.
        workers (int, optional): The number of workers to use for
            parallel processing. Used for the eager iterator. Default is
            8.
        chunk_mult (int, optional): The multiplier for the number of tiles
            to process in a single batch. Used for the eager iterator.
            Default is 2.
        agnostic_nms (bool, optional): Whether to use agnostic NMS.
            Default is True.
        conf_thr (float, optional): The confidence threshold for the
            NMS. Default is 0.25.
        iou (float, optional): The IoU threshold for the NMS. Default is
            0.7.
        device (str | None, optional): The device to use for inference.
            Default is None, will use "cuda" if available, otherwise
            "cpu". You can specify values like "cuda:0", "0", or
            multiple devices separated by commas.

    Returns:
        dict: A dictionary containing the inference results.
            - gdf (geopandas.GeoDataFrame): The inference results as a
              GeoDataFrame.
            - mag (float): The magnification used for inference.
            - mm_px (float): The micrometers per pixel used for inference.
            - time (dict): A dictionary containing the time taken for
              inference.
                - total (float): The total time taken for inference.

    Raises:
        ValueError: If both mag and mm_px are provided.

    """
    start_time = perf_counter()

    if isinstance(model, str):
        model = ultralytics.YOLO(model)

    ts = large_image_source_openslide.open(wsi_fp)
    ts_metadata = ts.getMetadata()

    mm_x = ts_metadata["mm_x"]
    mm_y = ts_metadata["mm_y"]

    # Get the desired resolution.
    if mag is not None and mm_px is not None:
        raise ValueError("Only one of mag or mm_px can be provided.")
    if mag is None and mm_px is None:
        # Use the scan resolution, we use the x resolution.
        mag, mm_px = return_mag_and_resolution(mm_px=mm_x)
    else:
        mag, mm_px = return_mag_and_resolution(mag=mag, mm_px=mm_px)

    # Calculate the x and y size of the tile at scan resolution.
    # desired resolution x sf_* -> scan resolution
    sf_x = mm_px / mm_x
    sf_y = mm_px / mm_y

    scan_tile_x = int(tile_size * sf_x)
    scan_tile_y = int(tile_size * sf_y)

    iterator = ts.eagerIterator(
        scale={"mm_x": mm_px, "mm_y": mm_px},
        tile_size={"width": tile_size, "height": tile_size},
        tile_overlap={"x": overlap, "y": overlap},
        chunk_mult=chunk_mult,
        batch=batch_size,
        prefetch=prefetch,
        workers=workers,
    )

    # Iterate through the WSI.
    boxes = []

    for i, batch in enumerate(iterator, start=1):
        print(f"\rProcessing batch {i}...    ", end="")
        tiles = batch["tile"].view()  # images are in BCHW format, as numpy

        # Convert a list of numpys, and change from RGB to BGR.
        tiles = [cv.cvtColor(tile, cv.COLOR_RGB2BGR) for tile in tiles]
        results = model(
            tiles,
            imgsz=tile_size,
            batch=batch_size,
            agnostic_nms=agnostic_nms,
            verbose=False,
            conf=conf_thr,
            iou=iou,
            device=device,
        )

        # Tile location at scan resolution.
        tile_x_coords = batch["gx"]
        tile_y_coords = batch["gy"]

        # Loop through tile variables.
        for result, x, y in zip(results, tile_x_coords, tile_y_coords):
            xyxys = result.boxes.xyxy
            cls_list = result.boxes.cls
            conf_list = result.boxes.conf

            for xyxy, cls, conf in zip(xyxys, cls_list, conf_list):
                # cls and conf are tensorts in device, convert to to int and float
                cls = int(cls)
                conf = float(conf)

                x1, y1, x2, y2 = xyxy

                # (1) scale to scan resolution, (2) shift to tile location.
                x1 = int(x1 * sf_x) + x
                y1 = int(y1 * sf_y) + y
                x2 = int(x2 * sf_x) + x
                y2 = int(y2 * sf_y) + y

                # Create the polygon.
                geom = Polygon([(x1, y1), (x2, y1), (x2, y2), (x1, y2)])
                boxes.append([cls, x1, y1, x2, y2, conf, geom])

    gdf = gpd.GeoDataFrame(
        boxes, columns=["label", "x1", "y1", "x2", "y2", "conf", "geometry"]
    )
    gdf["box_area"] = gdf.geometry.area

    gdf = non_max_suppression(gdf, conf_thr)
    gdf = remove_contained_boxes(gdf, iou).reset_index(drop=True)

    return {
        "gdf": gdf,
        "mag": mag,
        "mm_px": mm_px,
        "time": {"total": perf_counter() - start_time},
    }


def yolo_inference_on_region(
    wsi_fp: str,
    left: int,
    top: int,
    right: int,
    bottom: int,
    weights_fp: str,
    mag: float | None = 20,
    mm_px: float | None = None,
    tile_size: int = 640,
    stride: int | None = 480,
    batch_size: int = 64,
    agnostic_nms: bool = True,
    conf_thr: float = 0.25,
    iou: float = 0.7,
    device: str | None = None,
    pad_rgb: tuple[int, int, int] = (114, 114, 114),
    tile_area_threshold: float = 0.25,
    nms_iou_thr: float = 0.25,
    remove_contained_thr: float = 0.7,
) -> tuple[gpd.GeoDataFrame | pd.DataFrame, float, float]:
    """Inference on a region of a WSI using a YOLO model.

    Args:
        wsi_fp (str): The file path to the WSI.
        left (int): The left coordinate of the region.
        top (int): The top coordinate of the region.
        right (int): The right coordinate of the region.
        bottom (int): The bottom coordinate of the region.
        weights_fp (str): The file path to the YOLO model weights.
        mag (float | None, optional): The magnification to use for the
            inference. Default is 20.
        mm_px (float | None, optional): The pixel size to use for the
            inference. Default is None.
        tile_size (int, optional): The size of the tiles to use for the
            inference. This is the size of the tiles at the desired
            resolution. Default is 640.
        stride (int, optional): The stride to use for the inference. If
            None, then the stride will be the tile size (no overlap). This
            is the stride at the desired resolution. Default is 480.
        batch_size (int, optional): The batch size to use for the
            inference. Default is 64.
        agnostic_nms (bool, optional): Whether to use agnostic NMS.
            Default is True.
        conf_thr (float, optional): The confidence threshold to use for
            the inference. Default is 0.25.
        iou (float, optional): The IoU threshold to use for the
            inference. Default is 0.7.
        device (str | None, optional): The device to use for the
            inference. Default is None.
        pad_rgb (tuple[int, int, int], optional): The RGB values to pad
            the tiles with. Default is (114, 114, 114).
        tile_area_threshold (float, optional): The threshold for the tile
            area in region. Default is 0.25.
        nms_iou_thr (float, optional): The IoU threshold to use for the
            NMS after inference. Default is 0.25.
        remove_contained_thr (float, optional): The threshold for the
            contained boxes. Default is 0.7.

    Returns:
        tuple[gpd.GeoDataFrame, float, float]: A tuple containing the
            inference results, the magnification, and the pixel size used.

    Raises:
        ValueError: If the region is out of bounds.
    """
    assert left < right and top < bottom, "Region is out of bounds."

    if stride is None:
        stride = tile_size

    # Load the YOLO model.
    model = ultralytics.YOLO(weights_fp)

    ts = large_image_source_openslide.open(wsi_fp)
    ts_metadata = ts.getMetadata()

    mm_x = ts_metadata["mm_x"]
    mm_y = ts_metadata["mm_y"]

    # Get the desired resolution.
    if mag is not None and mm_px is not None:
        raise ValueError("Only one of mag or mm_px can be provided.")
    if mag is None and mm_px is None:
        # Use the scan resolution, we use the x resolution.
        mag, mm_px = return_mag_and_resolution(mm_px=mm_x)
    else:
        mag, mm_px = return_mag_and_resolution(mag=mag, mm_px=mm_px)

    # Calculate the x and y size of the tile at scan resolution.
    # desired resolution x sf_* -> scan resolution
    sf_x = mm_px / mm_x
    sf_y = mm_px / mm_y

    scan_tile_x = int(tile_size * sf_x)
    scan_tile_y = int(tile_size * sf_y)
    scan_stride_x = int(stride * sf_x)
    scan_stride_y = int(stride * sf_y)

    # Create a low res mask for the region.
    wsi_w = ts_metadata["sizeX"]
    wsi_h = ts_metadata["sizeY"]

    if right > wsi_w or bottom > wsi_h:
        raise ValueError("Region is out of bounds.")

    # Calculate the x, y coordinates
    xys = []

    for x in range(left, right, scan_stride_x):
        for y in range(top, bottom, scan_stride_y):
            xys.append((x, y))

    if len(xys) == 0:
        raise ValueError("No tiles to process with given region.")

    # Process in batches.
    batch_indices = np.arange(0, len(xys), batch_size)

    if len(batch_indices) == 1:
        print(
            f"Found {len(xys)} tiles to process in {len(batch_indices)} batch."
        )
    else:
        print(
            f"Found {len(xys)} tiles to process in {len(batch_indices)} batches."
        )

    # Iterate through the WSI.
    boxes = []

    # Calculate the tile area to be above the threshold.
    tile_area = tile_size * tile_size
    minimum_tile_area = tile_area * tile_area_threshold

    for batch_idx in batch_indices:
        xy_index = batch_idx * batch_size

        # Get the tiles.
        batch_xys = xys[xy_index : xy_index + batch_size]

        tiles = []

        tile_x_coords = []
        tile_y_coords = []

        for xy in batch_xys:
            x1, y1 = xy

            x2 = x1 + scan_tile_x

            if x2 > right:
                x2 = right

            y2 = y1 + scan_tile_y

            if y2 > bottom:
                y2 = bottom

            # Get the tiles.
            tile = ts.getRegion(
                region={
                    "left": x1,
                    "top": y1,
                    "right": x2,
                    "bottom": y2,
                },
                format="numpy",
                scale={"mm_x": mm_px, "mm_y": mm_px},
            )[0][:, :, :3].copy()

            # Pad the tile if it is not the right shape.
            tile_h, tile_w, _ = tile.shape

            tile_area = tile_h * tile_w

            if tile_area < minimum_tile_area:
                continue

            tile_x_coords.append(x1)
            tile_y_coords.append(y1)

            if tile_h != tile_size or tile_w != tile_size:
                tile = cv.copyMakeBorder(
                    tile,
                    0,
                    tile_size - tile_h,
                    0,
                    tile_size - tile_w,
                    cv.BORDER_CONSTANT,
                    value=pad_rgb,
                )
            # Convert to BGR.
            tile = cv.cvtColor(tile, cv.COLOR_RGB2BGR)

            tiles.append(tile)

        if len(tiles) == 0:
            continue

        results = model(
            tiles,
            imgsz=tile_size,
            batch=batch_size,
            agnostic_nms=agnostic_nms,
            verbose=False,
            conf=conf_thr,
            iou=iou,
            device=device,
        )

        # Loop through tile variables.
        for result, x, y in zip(results, tile_x_coords, tile_y_coords):
            xyxys = result.boxes.xyxy
            cls_list = result.boxes.cls
            conf_list = result.boxes.conf

            for xyxy, cls, conf in zip(xyxys, cls_list, conf_list):
                # cls and conf are tensorts in device, convert to to int and float
                cls = int(cls)
                conf = float(conf)

                x1, y1, x2, y2 = xyxy

                # (1) scale to scan resolution, (2) shift to tile location.
                x1 = int(x1 * sf_x) + x
                y1 = int(y1 * sf_y) + y
                x2 = int(x2 * sf_x) + x
                y2 = int(y2 * sf_y) + y

                # Create the polygon.
                geom = Polygon([(x1, y1), (x2, y1), (x2, y2), (x1, y2)])
                boxes.append([cls, x1, y1, x2, y2, conf, geom])

    gdf = gpd.GeoDataFrame(
        boxes, columns=["label", "x1", "y1", "x2", "y2", "conf", "geometry"]
    )
    gdf["box_area"] = gdf.geometry.area

    # Clean up predictions further.
    if nms_iou_thr > 0:
        gdf = non_max_suppression(gdf, nms_iou_thr)
    if remove_contained_thr > 0:
        gdf = remove_contained_boxes(gdf, remove_contained_thr)

    return gdf, mag, mm_px



"""Module for working with the girder client, or related to girder
documents, such as annotation documents.

Functions:
- login: Authenticate a girder client with the given credentials or
  interactively
- get_item_large_image_metadata: Get large image metadata for an item
- get_thumbnail: Get the thumbnail image by a specific magnification or
  shape
- get_region: Get image region from DSA
- get_element_contours: Get the contours of an element, regardless of
  the type
- get_roi_images: Get regions of interest (ROIs) as images from DSA
  annotations
- post_annotation: Post a new annotation to the DSA
- post_annotations_from_gdf: Post annotations from a GeoDataFrame
- remove_overlapping_annotations: Remove overlapping regions from
  elements
- upload_dir_to_dsa: Upload a local directory to a DSA item
- is_valid_color: Check if a string is a valid color string for DSA
  annotations.
- semantic_segmentation_annotation_metrics: Calculate metrics for
  semantic segmentation annotations.
- calculate_dice_from_annotations: Calculate the DICE score from two
  annotation documents.

"""

from girder_client import GirderClient, HttpError
import pickle, shutil, tempfile, re
import numpy as np
import pandas as pd
from pathlib import Path
import cv2 as cv
from copy import deepcopy
import geopandas as gpd
from shapely.geometry import Polygon, box, LineString
from .gpd_utils import remove_gdf_overlaps, make_gpd_valid

STANDARD_MM_PER_PX = 0.0002519
STANDARD_MAG = 40


def get_item_large_image_metadata(gc: GirderClient, item_id: str) -> dict:
    """Get large image metadata for an item.

    Args:
        gc (girder_client.GirderClient): The authenticated girder client.
        item_id (str): The item id.

    Returns:
        dict: The metadata of the large image item.

    """
    return gc.get(f"item/{item_id}/tiles")


def get_thumbnail(
    gc: GirderClient,
    item_id: str,
    mag: float | None = None,
    width: int | None = None,
    height: int | None = None,
    fill: int | tuple = (255, 255, 255),
) -> np.ndarray:
    """Get the thumbnail image by a specific magnification or shape. If mag is
    not None, then width and height are ignored. Fill is only used when both
    width and height are provided, to return the thumbnail at the exact shape.
    DSA convention will fill the height of the image, centering the image and
    filling the top and bottom of the image equally.

    Args:
        gc (girder_client.GirderClient): The authenticated girder client.
        item_id (str): The item id.
        mag (float, optional): The magnification. Defaults to None.
        width (int, optional): The width of the thumbnail. Defaults to None.
        height (int, optional): The height of the thumbnail. Defaults to None.
        fill (int | tuple, optional): The fill color. Defaults to (255, 255, 255).

    Returns:
        np.ndarray: The thumbnail image.

    """
    get_url = f"item/{item_id}/tiles/"

    if mag is not None:
        get_url += f"region?magnification={mag}&encoding=pickle"
    else:
        # Instead use width and height.
        params = ["encoding=pickle"]

        if width is not None and height is not None:
            if isinstance(fill, (tuple, list)):
                if len(fill) == 3:
                    fill = f"rgb({fill[0]},{fill[1]},{fill[2]})"
                elif len(fill) == 4:
                    fill = f"rgba({fill[0]},{fill[1]},{fill[2]},{fill[3]})"

            params.extend(
                [f"width={width}", f"height={height}", f"fill={fill}"]
            )
        elif width is not None:
            params.append(f"width={width}")
        elif height is not None:
            params.append(f"height={height}")

        get_url += "thumbnail?" + "&".join(params)

    response = gc.get(get_url, jsonResp=False)

    return pickle.loads(response.content)


def get_region(
    gc: GirderClient,
    item_id: str,
    left: int,
    top: int,
    width: int,
    height: int,
    mag: float | None = None,
) -> np.ndarray:
    """Get image region from DSA. Note that the output image might not
    be in the shape (width, height) if left + width or top + height
    exceeds the image size.

    Args:
        gc (girder_client.GirderClient): The authenticated girder client.
        item_id (str): The item id.
        left (int): The left coordinate.
        top (int): The top coordinate.
        width (int): The width of the region.
        height (int): The height of the region.
        mag (float, optional): The magnification. Defaults to None which
            returns the image at the scan magnification.

    Returns:
        np.ndarray: The region of the image.

    """
    get_url = (
        f"item/{item_id}/tiles/region?left={left}&top={top}&regionWidth="
        f"{width}&regionHeight={height}&encoding=pickle"
    )

    if mag is not None:
        get_url += f"&magnification={mag}"

    response = gc.get(get_url, jsonResp=False)

    return pickle.loads(response.content)


def get_element_contours(element: dict) -> np.ndarray:
    """Get the contours of an element, regardless of the type.

    Args:
        element (dict): The element dictionary.

    Returns:
        np.ndarray: The contours of the element.

    """
    if element["type"] == "rectangle":
        return get_rectangle_element_coords(element)
    else:
        return None


def get_roi_images(
    gc: GirderClient,
    item_id: str,
    save_dir: str,
    roi_groups: str | list,
    doc_names: str | list | None = None,
    mag: float | None = None,
    rgb_pad: tuple[int, int, int] | None = None,
) -> pd.DataFrame:
    """Gets regions of interest (ROIs) as images from DSA annotations.

    Args:
        gc (girder_client.GirderClient): The authenticated girder client.
        item_id (str): The item id.
        save_dir (str): The directory to save the roi images.
        roi_groups (str | list): The roi group name or list of roi group names.
        doc_names (str | list, optional): List of documents to look for, if None
            then it looks at all documents. Defaults to None.
        mag (float, optional): The magnification to get the roi images. Defaults
            to None which returns the images at scan magnification.
        rgb_pad (tuple[int, int, int], optional): The RGB values to pad the image with.
            This only is used when the annotation is a rotated rectangle or a polygon.

    Returns:
        pd.DataFrame: The roi images metadata.

    """
    if isinstance(roi_groups, str):
        roi_groups = [roi_groups]

    if isinstance(doc_names, str):
        doc_names = [doc_names]

    # Convert to a comma separated string.
    roi_groups_str = ",".join(roi_groups)

    if doc_names is None:
        docs = gc.get(
            f"annotation?itemId={item_id}&text="
            f"{roi_groups_str}&limit=0&offset=0&sort=lowerName&sortdir=1"
        )
    else:
        docs = []

        for doc_name in doc_names:
            docs.extend(
                gc.get(
                    f"annotation?itemId={item_id}&text={roi_groups_str}&name={doc_name}&limit=0&offset=0&sort=lowerName&sortdir=1"
                )
            )

    # Create the location to save the images.
    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)

    # Loop through each document.
    for doc in docs:
        # Get the full annotations.
        doc = gc.get(f"annotation/{doc['_id']}")

        # Get the elements.
        for element in doc.get("annotation", {}).get("elements", []):
            if element["group"] in roi_groups:
                # The annotations need to be handled based on type.
                contour = get_element_contours(element)

                # Get the minimum and maximum coordinates.
                xmin, ymin = contour.min(axis=0)
                xmax, ymax = contour.max(axis=0)
                w, h = xmax - xmin, ymax - ymin

                # Get the region of interest from the image.
                roi_image = get_region(gc, item_id, xmin, ymin, w, h, mag=mag)[
                    :, :, :3
                ]  # for now assuming images are RGB

                # Use pad if needed.
                if rgb_pad is not None:
                    roi_mask = np.ones((h, w), dtype=np.uint8)

                    # Draw the contours.
                    roi_mask = cv.drawContours(
                        roi_mask, [contour - (xmin, ymin)], -1, 0, cv.FILLED
                    )

                    roi_image[roi_mask == 1] = rgb_pad

                return roi_image, contour - (xmin, ymin)


def _rotate_point_list(point_list, rotation, center=(0, 0)):
    """Rotate a list of x, y points around a center location.
    Adapted from: https://github.com/DigitalSlideArchive/HistomicsTK/blob/master/histomicstk/annotations_and_masks/annotation_and_mask_utils.py
    INPUTS
    ------
    point_list : list
        list of x, y coordinates
    rotation : int or float
        rotation in radians
    center : list
        x, y location of center of rotation
    RETURN
    ------
    point_list_rotated : list
        list of x, y coordinates after rotation around center
    """
    point_list_rotated = []

    for point in point_list:
        cos, sin = np.cos(rotation), np.sin(rotation)
        x = point[0] - center[0]
        y = point[1] - center[1]

        point_list_rotated.append(
            (
                int(x * cos - y * sin + center[0]),
                int(x * sin + y * cos + center[1]),
            )
        )

    return point_list_rotated


def get_rectangle_element_coords(element: dict) -> np.ndarray:
    """Get the corner coordinate from a rectangle HistomicsUI element, can handle rotated elements.
    Adapted from: https://github.com/DigitalSlideArchive/HistomicsTK/blob/master/histomicstk/annotations_and_masks/annotation_and_mask_utils.py

    Args:
        element (dict): rectangle element from HistomicsUI annotation.

    Returns:
    corner_coords (np.ndarray): 4x2 array of corner coordinates of rectangle.

    """
    # element is a dict so prevent referencing
    element = deepcopy(element)

    # calculate the corner coordinates, assuming no rotation
    center_x, center_y = element["center"][:2]
    h, w = element["height"], element["width"]
    x_min = center_x - w // 2
    x_max = center_x + w // 2
    y_min = center_y - h // 2
    y_max = center_y + h // 2
    corner_coords = [
        (x_min, y_min),
        (x_max, y_min),
        (x_max, y_max),
        (x_min, y_max),
    ]

    # if there is rotation rotate
    if element["rotation"]:
        corner_coords = _rotate_point_list(
            corner_coords,
            rotation=element["rotation"],
            center=(center_x, center_y),
        )

    corner_coords = np.array(corner_coords, dtype=np.int32)

    return corner_coords


def login(
    api_url: str,
    login_or_email: str | None = None,
    password: str | None = None,
    api_key: str | None = None,
) -> GirderClient:
    """Authenticate a girder client with the given credentials or interactively
    if none is given.

    Args:
        api_url (str): The DSA girder API url.
        login_or_email (str | None): The login or email. Defaults to None.
        password (str | None): Password for login / email. Defaults to None.
        api_key (str | None): The api key to authenticate with. Defaults to None.

    Returns:
        girder_client.GirderClient: The authenticated girder client.

    """
    gc = GirderClient(apiUrl=api_url)

    if api_key is None:
        if login_or_email is None:
            _ = gc.authenticate(interactive=True)
        elif password is None:
            _ = gc.authenticate(username=login_or_email, interactive=True)
        else:
            _ = gc.authenticate(username=login_or_email, password=password)
    else:
        _ = gc.authenticate(apiKey=api_key)

    return gc


def get_items(gc: GirderClient, parend_id: str) -> list[dict]:
    """Get the items in a parent location recursively.

    Args:
        gc (girder_client.GirderClient): The authenticated girder client.
        parend_id (str): The parent id to start the search (folder / collection).

    Returns:
        list[dict]: The list of items.

    """
    params = {
        "type": "folder",
        "limit": 0,
        "offset": 0,
        "sort": "_id",
        "sortdir": 1,
    }

    request_url = f"resource/{parend_id}/items"

    try:
        items = gc.get(request_url, parameters=params)
    except HttpError:
        params["type"] = "collection"

        items = gc.get(request_url, parameters=params)

    return items


def get_roi_with_yolo_labels_from_single_doc(
    gc: GirderClient,
    ann_doc: dict,
    roi_element_group: str,
    group_map: dict[str, int],
    mag: float | None = None,
    rgb_fill: tuple[int, int, int] = (114, 114, 114),
) -> tuple[np.ndarray, str]:
    """Get the ROI image with the YOLO labels given a DSA annotation document.

    Args:
        gc (girder_client.GirderClient): Authenticated girder client.
        ann_doc (dict): Annotation document metadata.
        roi_element_group (str): The group name of the ROI element.
        group_map (dict[str, int]): Mapping of group names to group IDs.
        mag (float, optional): The magnification to use for the ROI. Defaults to
            None which uses the default magnification.
        rgb_fill (tuple[int, int, int], optional): The RGB color to fill the ROI
            with. Defaults to (114, 114, 114).

    Returns:
        tuple[np.ndarray, str]: The ROI image and YOLO labels data as string.

    Raises:
        ValueError: If multiple ROI elements are found in the document.
        ValueError: If ROI element type is not polyline.

    """
    roi_element = None
    box_elements = []

    # Loop through all elements in the annotation document.
    for element in ann_doc.get("annotation", {}).get("elements", []):
        if element.get("group") == roi_element_group:
            # Append the ROI element, there can't more than one.
            if roi_element is not None:
                raise ValueError(
                    "Multiple ROI elements found in the document."
                )

            roi_element = element
        elif element.get("group") in group_map:
            # Append the box elements.
            box_elements.append(element)

    # Get the large image metadata.
    large_image_metadata = get_item_large_image_metadata(gc, ann_doc["itemId"])
    scan_mag = large_image_metadata["magnification"]

    # Specify magnification to use.
    if mag is None:
        mag = scan_mag

    # Calculate the factor to convert from scan mag to desired mag.
    sf = mag / scan_mag

    # Get the ROI image.
    if roi_element.get("type") == "polyline":
        # Grab the smallest bounding box.
        points = np.array(roi_element["points"])[:, :2]  # remove z-axis
        roi_x1, roi_y1 = points.min(axis=0)
        roi_x2, roi_y2 = points.max(axis=0)

        img = get_region(
            gc,
            ann_doc["itemId"],
            roi_x1,
            roi_y1,
            roi_x2 - roi_x1,
            roi_y2 - roi_y1,
            mag=mag,
        )[
            :, :, :3
        ]  # make sure it is RGB

        # Shift the points to be relative to smallest bounding box.
        points -= [roi_x1, roi_y1]

        # Create a mask and draw the ROI filled on it (handles rotated ROIs).
        mask = np.zeros(img.shape[:2], dtype=np.uint8)
        mask = cv.drawContours(
            mask, [(points * sf).astype(np.int32)], 0, 1, cv.FILLED
        )

        # Gray out regions outside of ROI.
        img[np.where(mask == 0)] = rgb_fill
    else:
        raise ValueError("Only polyline ROIs are supported.")

    roi_h, roi_w = img.shape[:2]

    # Convert the box elements into YOLO format.
    labels_data = ""

    for element in box_elements:
        if element.get("group") in group_map:
            if element.get("type") != "rectangle":
                print(
                    f"Skipping element: {element.get('type')} not supported."
                )
                continue

            group_id = group_map[element.get("group")]

            # Get the center of box and shift to be relative to ROI.
            x_center, y_center = (
                np.array(element["center"])[:2] - [roi_x1, roi_y1]
            ) * sf

            # Normalize the coordinates.
            x_center /= roi_w
            y_center /= roi_h

            # Scale the box width and height and then normalize.
            width, height = (
                element["width"] * sf / roi_w,
                element["height"] * sf / roi_h,
            )

            labels_data += f"{group_id} {x_center:.6f} {y_center:.6f} {width:.6f} {height:.6f}\n"
        else:
            print(
                f"Skipping element: group {element.get('group')} not found in group map."
            )

    # Return the ROI image and labels data.
    return img, labels_data.strip()


def post_annotations_from_gdf(
    gc: GirderClient,
    item_id: str,
    doc_name: str,
    gdf: gpd.GeoDataFrame,
    idx_config: dict,
    add_attr: dict | None = None,
) -> dict:
    """Post annotations from a GeoDataFrame to the DSA.

    Args:
        gc (girder_client.GirderClient): Girder client.
        item_id (str): DSA item id, annotation posted to this item.
        doc_name (str): Name new annotation document will have.
        gdf (GeoDataFrame): The GeoDataFrame with the polygons, includes
            label column.
        idx_config (dict): For each label in the GeoDataFrame, the
            configuration for the DSA annotation. Include "label",
            "fillColor", "lineColor", "lineWidth", and "group".
        add_attr (dict | None, optional): Additional attributes to add
            to the annotation. Defaults to None. Should include "modelId"
            and "modelName" if applicable.

    Returns:
        dict: The response from the DSA.

    """
    if add_attr is None:
        add_attr = {}

    elements = []

    # This step makes sure that each row contains a single polygon.
    # Multi-polygons will be exploded into multiple rows.
    gdf = gdf.explode(index_parts=False).reset_index(drop=True)

    # Process the dataframe.
    for _, r in gdf.iterrows():
        label = r["label"]
        poly = r["geometry"]

        exterior_poly = list(poly.exterior.coords)
        interior_polys = [list(interior.coords) for interior in poly.interiors]

        points = [[int(xy[0]), int(xy[1]), 0] for xy in exterior_poly]

        if len(points) == 0:
            continue

        holes = []

        for interior_poly in interior_polys:
            hole = [[int(xy[0]), int(xy[1]), 0] for xy in interior_poly]
            holes.append(hole)

        element = {
            "points": points,
            "fillColor": idx_config[label]["fillColor"],
            "lineColor": idx_config[label]["lineColor"],
            "type": "polyline",
            "lineWidth": 2,
            "closed": True,
            "label": {"value": idx_config[label]["label"]},
            "group": idx_config[label]["group"],
        }

        add_params = {}

        if len(holes):
            element["holes"] = holes

        elements.append(element)

    response = gc.post(
        "annotation",
        parameters={
            "itemId": item_id
        },  # "modelId": model_id, "modelName": model_name
        json={
            "name": doc_name,
            "description": "",
            "elements": elements,
            "attributes": add_attr,
            # **add_attr,
        },
    )

    return response


def get_thumbnail_with_mask(
    gc: GirderClient,
    item_id: str,
    mag: float,
    doc_name: str,
    label2id: dict,
    background_label: int = 0,
):
    """Get the thumbnail (i.e. low resolution image) and a label mask
    drawn from annotation documents for an item.

    Args:
        gc (girder_client.GirderClient): The Girder client to use.
        item_id (str): The item id of the image to get the thumbnail
            and mask from.
        mag (float): The magnification to get the thumbnail at.
        doc_name (str): The name of the annotation document(s) to use.
        label2id (dict): A dictionary mapping the label names to ids,
            for generating the mask.
        background_label (int, optional): The label to use for the
            background. Defaults to 0.

    Returns:
        tuple: A tuple containing the thumbnail image and the mask.

    """
    # Get the thumbnail of the image.
    thumbnail = get_thumbnail(gc, item_id, mag=mag)[:, :, :3]

    # Get the width and h eight of the thumbnail.
    h, w = thumbnail.shape[:2]

    # Get the item large image metadata to get its scan magnification.
    large_image_metadata = gc.get(f"item/{item_id}/tiles")
    scan_mag = large_image_metadata["magnification"]

    # Calculate a scale factor to go from full scale coordinates to thumbnail ones.
    sf = mag / scan_mag

    # Create a blank background mask.
    mask = np.ones((h, w), dtype=np.uint8) * background_label

    # Get annotation metadata for docs of interest.
    for ann_doc in gc.get(
        f"annotation",
        parameters={"itemId": item_id, "name": doc_name, "limit": 0},
    ):
        # Get the full annotation document.
        doc = gc.get(f"annotation/{ann_doc['_id']}")

        # Get the elements.
        for element in doc.get("annotation", {}).get("elements", []):
            # Check if the label of the element is in the label2id dictionary.
            if element.get("label", {}).get("value") not in label2id:
                continue

            label = element["label"]["value"]

            # Create a mask for this element as background.
            element_mask = np.ones((h, w), dtype=np.uint8) * background_label

            if element["type"] == "polyline":
                points = (np.array(element["points"])[:, :2] * sf).astype(int)

                holes = []

                for hole in element.get("holes", []):
                    holes.append((np.array(hole)[:, :2] * sf).astype(int))
            elif element["rectangle"]:
                # Format the rectangle as a polyline.
                points = (get_rectangle_element_coords(element) * sf).astype(
                    int
                )

            # Draw the points on the mask with 1.
            element_mask = cv.drawContours(
                element_mask, [points], -1, 1, cv.FILLED
            )

            if len(holes):
                # Draw the holes on the mask with 0.
                element_mask = cv.drawContours(
                    element_mask, holes, -1, 0, cv.FILLED
                )

            # Specify the label of the elements on the thumbnail mask.
            mask[element_mask == 1] = label2id[label]

    # Return the thumbnail and the mask.
    return thumbnail, mask


def remove_overlapping_annotations(
    annotation: dict,
    group_order: list[str],
    max_coords: tuple[int, int] | None = None,
) -> dict:
    """Remove overlapping regions from elements.

    annotation (dict): The annotation key from a DSA annotation
        document.
    group_order (list[str]): The order of groups to process. Groups
        earlier in the list will take precedence over those later when
        removing overlapping regions. Groups in the document that are
        not in the list will be kept as is.
    max_coords (tuple[int, int], optional): The maximum coordinates in
        the width and height dimension. If annotation element
        coordinates are larger than this, they will be clipped to the
        maximum coordinate. If None is passed then no clipping will be
        performed. Defaults to None.

    Returns:
        dict: The modified annotation document.

    """
    # Check if the elements key is present.
    if "elements" not in annotation:
        print("No elements key found in annotation.")

    elements = annotation.get("elements", [])

    new_elements = []
    elements_to_process = []

    # Loop through the elements.
    for element in elements:
        # Pop the id key.
        _ = element.pop("id", None)

        # Check if group is in the group order.
        if element.get("group") in group_order:
            # Process the element.
            if element.get("type") == "polyline":
                points = element.pop("points")

                # Convert the points to a x, y numpy array.
                points = np.array(
                    [[p[0], p[1]] for p in points], dtype=np.float32
                )

                # Add holes to the polygon if they exists.
                if "holes" in element and len(element["holes"]):
                    # Remove the holes key.
                    holes = element.pop("holes")

                    # Convert them to numpy array.
                    holes = [
                        np.array(
                            [[p[0], p[1]] for p in hole], dtype=np.float32
                        )
                        for hole in holes
                    ]
                else:
                    holes = None

                # Add the points as a geometry object.
                element["geometry"] = Polygon(points, holes=holes)

                # Add the points to the elements to process.
                elements_to_process.append(element)
            elif element.get("type") == "rectangle":
                # For ease of use, convert the rectangle to a polyline.
                coords = get_rectangle_element_coords(element)
                element["geometry"] = Polygon(coords)
                element["type"] = "polyline"

                # Remove the keys that are no longer needed.
                _ = element.pop("rotation", None)
                _ = element.pop("width", None)
                _ = element.pop("height", None)
                _ = element.pop("center", None)

                # Add the points to the elements to process.
                elements_to_process.append(element)
            else:
                # Keep the element as is.
                new_elements.append(element)
        else:
            # Keep the element as is.
            new_elements.append(element)

    # Convert the elements to process into a GeoDataFrame.
    if len(elements_to_process):
        gdf = gpd.GeoDataFrame(elements_to_process)

        if max_coords is not None:
            # Create a bounding box polygon.
            bbox = box(0, 0, max_coords[0], max_coords[1])

            # Clip the polygons to the bounding box.
            gdf = gdf.clip(bbox)

            # This could have created multiple polygons, explode them.
            gdf = gdf.explode(index_parts=False).reset_index(drop=True)

        # Map by the groups.
        group_map = {group: i for i, group in enumerate(group_order)}

        # Add the order column, set by the group map.
        gdf["order"] = gdf["group"].map(group_map)

        # Remove the overlapping polygons.
        gdf = remove_gdf_overlaps(gdf, "order")

        # Convert the GeoDataFrame back to a list of dictionaries.
        elements_to_process = gdf.to_dict(orient="records")

        # Add the elements back to the list.
        for element in elements_to_process:
            # Convert the geometry back to the original format.
            del element["order"]
            geometry = element.pop("geometry")

            # If the object ended up being a line, skipt it.
            if isinstance(geometry, LineString):
                continue

            exterior_poly = list(geometry.exterior.coords)
            interior_polys = [
                list(interior.coords) for interior in geometry.interiors
            ]

            points = [[int(xy[0]), int(xy[1]), 0] for xy in exterior_poly]

            if len(points) == 0:
                continue

            holes = []

            for interior_poly in interior_polys:
                hole = [[int(xy[0]), int(xy[1]), 0] for xy in interior_poly]
                holes.append(hole)

            element["points"] = points
            element["holes"] = holes

            new_elements.append(element)

    annotation["elements"] = new_elements

    # Return the annotation.
    return annotation


def post_annotation(gc: GirderClient, item_id: str, annotation: dict) -> dict:
    """Post a new annotation to the DSA.

    Args:
        gc (girder_client.GirderClient): The authenticated girder
            client.
        item_id (str): The item id to post the annotation to.
        annotation (dict): The annotation dictionary post, it should
            include only the "name", "description", and "elements" keys.

    Returns:
        dict: The response from the DSA.

    Raises:
        ValueError: If the annotation dictionary does not include the
            "name" or "elements" keys.
        ValueError: If the annotation dictionary includes any keys other
            than "name", "description", "attributes", "display", and
            "elements".

    """
    if "name" not in annotation or "elements" not in annotation:
        raise ValueError("Annotation must include 'name' and 'elements' keys.")

    if "description" not in annotation:
        annotation["description"] = ""

    for k in annotation.keys():
        if k not in [
            "name",
            "description",
            "attributes",
            "display",
            "elements",
        ]:
            raise ValueError(
                f'Annotation key "{k}" is not allowed. Only "name", "description", "attributes", "display", and "elements" are allowed.'
            )

    return gc.post(
        "/annotation",
        parameters={"itemId": item_id},
        json=annotation,
    )


def upload_dir_to_dsa(
    gc: GirderClient,
    checkpoint_dir: str,
    item_id: str,
):
    """Upload a local directory to a DSA item.

    Args:
        gc (girder_client.GirderClient): The girder client.
        checkpoint_dir (str): The path to the local directory to upload.
        item_id (str): The ID of the DSA item to upload the directory to.

    """
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)

        # Zip the directory.
        shutil.make_archive(temp_dir / "checkpoint", "zip", checkpoint_dir)

        # Upload the zip file to the DSA item.
        _ = gc.uploadFileToItem(
            item_id,
            str(temp_dir / "checkpoint.zip"),
        )


def is_valid_color(color_string: str) -> bool:
    """Check if a string is a valid color string for DSA annotations.

    Args:
        color_string (str): The color string to check.

    Returns:
        bool: True if the color string is valid, False otherwise.

    """
    # Separate patterns for RGB and RGBA
    rgb_pattern = r"^rgb\(\s*(\d{1,3})\s*,\s*(\d{1,3})\s*,\s*(\d{1,3})\s*\)$"
    rgba_pattern = r"^rgba\(\s*(\d{1,3})\s*,\s*(\d{1,3})\s*,\s*(\d{1,3})\s*,\s*([01]\.?\d*)\s*\)$"

    def is_valid_rgb(color_string):
        match = re.match(rgb_pattern, color_string)
        if not match:
            return False

        # Extract the captured groups (the numbers)
        r, g, b = map(int, match.groups())

        # Validate the ranges
        return 0 <= r <= 255 and 0 <= g <= 255 and 0 <= b <= 255

    def is_valid_rgba(color_string):
        match = re.match(rgba_pattern, color_string)
        if not match:
            return False

        # Extract the captured groups
        r, g, b, a = match.groups()

        # Convert to numbers
        r, g, b = map(int, [r, g, b])
        a = float(a)

        # Validate RGB values (0-255)
        if not (0 <= r <= 255 and 0 <= g <= 255 and 0 <= b <= 255):
            return False

        # Validate alpha (0-1)
        if not (0 <= a <= 1):
            return False

        return True

    if is_valid_rgb(color_string):
        return True
    if is_valid_rgba(color_string):
        return True
    return False


def semantic_segmentation_annotation_metrics(
    gc: GirderClient,
    item_id: str,
    gt_ann_docname: str,
    inf_ann_docname: str,
    ignore_groups: None | list[str] = None,
    include_groups: None | list[str] = None,
    gt_label_to_col: bool = False,
    inf_label_to_col: bool = False,
) -> dict:
    """
    Calculate the metrics for the test dataset for a given annotation
    document name.

    Args:
        gc (girder_client.GirderClient): Authenticated girder client.
        item_id (str): The item id of the WSI.
        gt_ann_docname (str): The name of the ground truth annotation
            document.
        inf_ann_docname (str): The name of the inference annotation
            document.
        ignore_groups (list[str] | None, optional): A list of groups to
            ignore. Defaults to None.
        include_groups (list[str] | None, optional): A list of groups to
            include. Defaults to None.
        gt_label_to_col (bool, optional): Whether to convert the ground
            truth label column to the group column. Defaults to False.
        inf_label_to_col (bool, optional): Whether to convert the
            inference label column to the group column. Defaults to
            False.

    Raises:
        ValueError: If multiple annotation documents of the same name
            are found.
        ValueError: If no annotation document of the given name is
            found.

    Returns:
        dict: A dictionary containing the metrics.

    """
    # Get the annotation documents.
    ann_docs = gc.get(
        "annotation",
        parameters={"itemId": item_id, "name": gt_ann_docname, "limit": 0},
    )

    if len(ann_docs) > 1:
        raise ValueError(
            f"Multiple annotation documents of name {gt_ann_docname} found. Unsure which to use."
        )
    elif len(ann_docs):
        # Get the full annotation document in geojson format.
        gt_ann_doc = gc.get(f"annotation/{ann_docs[0]['_id']}/geojson")
    else:
        raise ValueError(
            f"No annotation document of name {gt_ann_docname} found."
        )

    ann_docs = gc.get(
        "annotation",
        parameters={"itemId": item_id, "name": inf_ann_docname, "limit": 0},
    )

    if len(ann_docs) > 1:
        raise ValueError(
            f"Multiple annotation documents of name {inf_ann_docname} found. Unsure which to use."
        )
    elif len(ann_docs):
        # Get the full annotation document in geojson format.
        inf_ann_doc = gc.get(f"annotation/{ann_docs[0]['_id']}/geojson")
    else:
        raise ValueError(
            f"No annotation document of name {inf_ann_docname} found."
        )

    # Convert the geojson to a geopandas dataframe.
    gt_gdf = gpd.GeoDataFrame.from_features(gt_ann_doc["features"])
    inf_gdf = gpd.GeoDataFrame.from_features(inf_ann_doc["features"])

    # Convert the label to a column if requested.
    if gt_label_to_col:
        gt_gdf["group"] = gt_gdf["label"].apply(
            lambda x: x.get("value", "") if isinstance(x, dict) else x
        )

    if inf_label_to_col:
        inf_gdf["group"] = inf_gdf["label"].apply(
            lambda x: x.get("value", "") if isinstance(x, dict) else x
        )

    if ignore_groups is not None and len(ignore_groups) == 0:
        ignore_groups = None

    if include_groups is not None and len(include_groups) == 0:
        include_groups = None

    if ignore_groups is not None:
        # Get a dataframe of ignore ground truth.
        ignore_gdf = gt_gdf[gt_gdf["group"].isin(ignore_groups)]

    if include_groups is None:
        # Get the unique groups in the ground truth.
        include_groups = gt_gdf["group"].unique().tolist()

    # Filter to only include the groups of interest.
    gt_gdf = gt_gdf[gt_gdf["group"].isin(include_groups)]
    inf_gdf = inf_gdf[inf_gdf["group"].isin(include_groups)]

    # Subtract the ignore ground truth from the inference.
    if ignore_groups is not None:
        inf_gdf = gpd.overlay(
            inf_gdf, ignore_gdf, how="difference", keep_geom_type=False
        )

    # Turns the gdfs into a single row per group, as multipolygons as needed.
    gt_gdf = gt_gdf.dissolve(by="group", as_index=False)

    if any(inf_gdf.is_valid == False):
        inf_gdf["geometry"] = inf_gdf["geometry"].make_valid()

    inf_gdf = inf_gdf.dissolve(by="group", as_index=False)

    # Iterate through each label.
    dice_scores = []
    class_weights = []

    for label in include_groups:
        # Filter each gdf to only include the label.
        gt_gdf_label = gt_gdf[gt_gdf["group"] == label]["geometry"]
        inf_gdf_label = inf_gdf[inf_gdf["group"] == label]["geometry"]

        if len(gt_gdf_label) == 0 or len(inf_gdf_label) == 0:
            intersection = 0
        else:
            intersection = gt_gdf_label.intersection(inf_gdf_label).area.sum()

        # Calculate the sum of the areas.
        gt_area = gt_gdf_label.area.sum()
        class_weights.append(gt_area)
        sum_of_areas = gt_area + inf_gdf_label.area.sum()

        if sum_of_areas:
            dice_scores.append(float(2 * intersection / sum_of_areas))
        else:
            dice_scores.append(1)

    dice_scores = np.array(dice_scores)
    class_weights = np.array(class_weights)

    # Calculate the weighted mean of the dice scores.
    weighted_mean_dice = np.sum(dice_scores * class_weights) / np.sum(
        class_weights
    )

    metrics = {"per_class_dice": {}}
    for label, dice_score in zip(include_groups, dice_scores):
        metrics["per_class_dice"][label] = float(dice_score)

    metrics["weighted_mean_dice"] = float(weighted_mean_dice)

    return metrics


def calculate_dice_from_annotations(
    gt_ann_doc: dict,
    inf_ann_doc: dict,
    classes: list[str],
) -> tuple[dict, float]:
    """
    Calculate the DICE score between two annotation documents.

    Args:
        gt_ann_doc (dict): The ground truth annotation document, in
            geojson format.
        inf_ann_doc (dict): The inference annotation document, in
            geojson format.
        classes (list[str]): Classes to calculate dice scores for.

    Returns:
        dict: A dictionary containing the DICE scores for each class.
        float: The weighted mean DICE score, weighted by the amount of
            area of each class in the ground truth.

    """
    gt_gdf = gpd.GeoDataFrame.from_features(gt_ann_doc["features"])
    inf_gdf = gpd.GeoDataFrame.from_features(inf_ann_doc["features"])

    gt_gdf["label"] = gt_gdf["label"].apply(lambda x: x["value"])
    inf_gdf["label"] = inf_gdf["label"].apply(lambda x: x["value"])

    gt_gdf = gt_gdf[gt_gdf["label"].isin(classes)]
    inf_gdf = inf_gdf[inf_gdf["label"].isin(classes)]

    gt_gdf = make_gpd_valid(gt_gdf)
    inf_gdf = make_gpd_valid(inf_gdf)

    # Turns the gdfs into a single row per group, as multipolygons as needed.
    gt_gdf = gt_gdf.dissolve(by="label", as_index=False)
    inf_gdf = inf_gdf.dissolve(by="label", as_index=False)

    # Iterate through each group.
    dice_scores = []
    class_weights = []

    for cls in classes:
        gt_gdf_group = gt_gdf[gt_gdf["label"] == cls]["geometry"].reset_index(
            drop=True
        )
        inf_gdf_group = inf_gdf[inf_gdf["label"] == cls][
            "geometry"
        ].reset_index(drop=True)

        if len(gt_gdf_group) == 0 or len(inf_gdf_group) == 0:
            intersection = 0
        else:
            intersection = gt_gdf_group.intersection(inf_gdf_group).area.sum()

        gt_area = gt_gdf_group.area.sum()
        class_weights.append(gt_area)

        sum_of_areas = gt_area + inf_gdf_group.area.sum()

        if sum_of_areas:
            dice_scores.append(float(2 * intersection / sum_of_areas))
        else:
            dice_scores.append(1)

    dice_scores = np.array(dice_scores)
    class_weights = np.array(class_weights)

    # Calculate the weighted mean of the dice scores.
    weighted_mean_dice = np.sum(dice_scores * class_weights) / np.sum(
        class_weights
    )

    dice_scores = {
        cls: float(score) for cls, score in zip(classes, dice_scores)
    }

    return dice_scores, float(weighted_mean_dice)


def get_mag_and_mm_px(
    ts_metadata: dict, mag: float | None = None, mm_px: float | None = None
) -> tuple[float, float]:
    """Get the magnification and micrometers per pixel from the tile source metadata.

    Args:
        ts_metadata: Metadata from the tile source.

    """
    # Calculate the mm per pixel to use.
    if mm_px is None and mag is None:
        print("mm_px and mag were not specified, using scan resolution.")
        mag_x = STANDARD_MAG * STANDARD_MM_PER_PX / ts_metadata["mm_x"]
        mag_y = STANDARD_MAG * STANDARD_MM_PER_PX / ts_metadata["mm_y"]
        mm_x = ts_metadata["mm_x"]
        mm_y = ts_metadata["mm_y"]
    elif mm_px is None:
        # Calculate the mm per pixel.
        mm_x = STANDARD_MAG * STANDARD_MM_PER_PX / mag
        mm_y = mm_x
        mag_x = mag_y = mag
    else:
        # Calculate the magnification.
        mag_x = STANDARD_MAG * STANDARD_MM_PER_PX / mm_px
        mag_y = mag_x
        mm_x = mm_y = mm_px

    # Scale factor, multiply to go from scan magnification to desired mag.
    sf_x = ts_metadata["mm_x"] / mm_x
    sf_y = ts_metadata["mm_y"] / mm_y

    return (mag_x, mag_y), (mm_x, mm_y), (sf_x, sf_y)



"""Utility functions for DSA Helpers.

This module provides various miscellaneous utility functions that are
not grouped into their own modules.

Function list:
* non_max_suppression: Apply non-max suppression (NMS) on a set of prediction boxes.

"""

from shapely.geometry import Polygon
import numpy as np
import cv2 as cv
import pandas as pd


def non_max_suppression(df: pd.DataFrame, thr: float) -> pd.DataFrame:
    """Apply non-max suppression (nms) on a set of prediction boxes.
    Source: https://github.com/rbgirshick/fast-rcnn/blob/master/lib/utils/nms.py

    Args
    ------
    df : pd.DataFrame
        Data for each box, must contain the x1, y1, x2, y2, conf columns
        with point 1 being top left of the box and point 2 and bottom
        right of box.
    thr (float): IoU threshold used for NMS.

    Returns:
        pd.DataFrame: Remaining boxes.

    """
    df = df.reset_index(drop=True)  # indices must be reset
    dets = df[["x1", "y1", "x2", "y2", "conf"]].to_numpy()
    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(ovr <= thr)[0]
        order = order[inds + 1]

    return df.loc[keep]


def return_mag_and_resolution(
    mag: float | int | None = None,
    mm_px: float | None = None,
    standard_mag: int = 40,
    standard_mm_px: float = 0.0002519,
) -> tuple[float, float]:
    """Given an input magnification or resolution, return the other
    based on the standard values defined by standard_mag and
    standard_mm_per_px.

    Args:
        mag (float | int | None, optional): Magnification to use. Defaults to
            None.
        mm_px (float | None, optional): Resolution to use. Defaults to
            None.
        standard_mag (int, optional): Standard magnification. Matches
            the standard resolution provided. Defaults to 40.
        standard_mm_px (float, optional): Standard resolution. Matches
            the standard magnification provided. Defaults to 0.0002519.

    Returns:
        tuple[float, float]: Returns the magnification and
            resolution of the image. These are standardized to the
            standard values provided in the inputs.

    """
    assert (
        mag is not None or mm_px is not None
    ), "Either mag or mm_px must be provided."
    assert (
        mag is None or mm_px is None
    ), "Only one of mag or mm_px can be provided."

    if mm_px is None:
        # Use the magnification to calculate the resolutions.
        mm_px = standard_mm_px * standard_mag / mag
    else:
        # Use the resolution to calculate the magnification.
        mag = standard_mag * standard_mm_px / mm_px

    return mag, mm_px


def remove_small_holes(
    polygon: Polygon, hole_area_threshold: float
) -> Polygon:
    """Remove small holes from a shapely polygon.

    Args:
        polygon (shapely.geometry.Polygon): Polygon to remove holes
            from.
        hole_area_threshold (float): Minimum area of a hole to keep it.

    Returns:
        shapely.geometry.Polygon: Polygon with small holes removed.

    """
    if not polygon.interiors:  # if there are no holes, return as is
        return polygon

    # Filter out small holes
    new_holes = [
        hole
        for hole in polygon.interiors
        if Polygon(hole).area > hole_area_threshold
    ]

    # Create a new polygon with only large holes
    return Polygon(polygon.exterior, new_holes)


def convert_to_json_serializable(
    data: int | float | str | list | dict,
) -> int | float | str | list | dict:
    """Convert a list, integer, float, or dictionary into a JSON
    serializable version of the object. Uses recursion to make sure the
    entire input data structure is converted to a JSON serializable
    version.

    Args:
        data (int | float | str | list | dict): Data to convert to
            JSON serializable.

    Returns:
        int | float | str | list | dict: JSON serializable version of
            the input data.

    """
    if isinstance(data, np.ndarray):
        return data.tolist()
    elif isinstance(data, np.integer):
        return int(data)
    elif isinstance(data, np.floating):
        return float(data)
    elif isinstance(data, dict):
        return {
            key: convert_to_json_serializable(value)
            for key, value in data.items()
        }
    elif isinstance(data, list):
        return [convert_to_json_serializable(item) for item in data]
    return data


def contours_to_polygons(binary_image: np.ndarray) -> list[Polygon]:
    """Convert a binary image to a list of polygons.

    Args:
        binary_image (numpy.ndarray): Binary image.

    Returns:
        list[shapely.geometry.Polygon]: List of polygons.

    """
    contours, hierarchy = cv.findContours(
        binary_image, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE
    )
    hierarchy = hierarchy[0]  # shape: (n_contours, 4)

    def contour_to_coords(contour):
        return [(int(p[0][0]), int(p[0][1])) for p in contour]

    polygons = []
    used = set()

    for i, (next_idx, prev_idx, child_idx, parent_idx) in enumerate(hierarchy):
        if parent_idx != -1:
            # Skip child contours for now (they'll be added as holes)
            continue

        # Outer contour
        exterior = contour_to_coords(contours[i])

        if len(exterior) < 4:
            # Skip contours that are too small
            used.add(i)
            continue

        holes = []

        # Look for children (holes)
        child = hierarchy[i][2]
        while child != -1:
            hole_coords = contour_to_coords(contours[child])
            holes.append(hole_coords)
            used.add(child)
            # Check for nested children (e.g. box in a hole in a box)
            grandchild = hierarchy[child][2]
            if grandchild != -1:
                # Treat grandchild as a new polygon later
                pass
            child = hierarchy[child][0]  # next sibling hole

        used.add(i)
        polygon = Polygon(exterior, holes)

        if not polygon.is_valid:
            # Attempt to fix the polygon
            polygon = polygon.buffer(0)

        if polygon.is_valid:
            polygons.append(polygon)

    # Add any unused outer contours (e.g. nested objects)
    for i in range(len(contours)):
        if i not in used:
            coords = contour_to_coords(contours[i])

            if len(coords) < 4:
                # Skip contours that are too small
                continue

            poly = Polygon(coords)

            if not poly.is_valid:
                # Attempt to fix the polygon
                poly = poly.buffer(0)

            if poly.is_valid:
                polygons.append(poly)

    return polygons



import torch
import geopandas as gpd
import histomicstk as htk
import numpy as np
import large_image_source_openslide
from PIL import Image
from time import perf_counter, sleep
from tqdm import tqdm
from multiprocessing import Pool
from transformers import (
    SegformerForSemanticSegmentation,
    SegformerImageProcessor,
)

from shapely import make_valid
from shapely.affinity import scale
from shapely.geometry import Polygon
from shapely.ops import unary_union

from ...image_utils import label_mask_to_polygons
from ...gpd_utils import rdp_by_fraction_of_max_dimension, make_multi_polygons
from ..inference_results import InferenceResult

stain_color_map = htk.preprocessing.color_deconvolution.stain_color_map

# specify stains of input image
stains = [
    "hematoxylin",  # nuclei stain
    "eosin",  # cytoplasm stain
    "null",
]  # set to null if input contains only two stains

# create stain matrix
W = np.array([stain_color_map[st] for st in stains]).T

# The standard values are taken from Emory's 40x scanned images.
STANDARD_MM_PER_PX = 0.0002519
STANDARD_MAG = 40


def inference(
    model: str | torch.nn.Module,
    wsi_fp: str,
    label_ranks: list[int] | None = None,
    batch_size: int = 64,
    tile_size: int = 512,
    mag: float | None = 10.0,
    mm_px: float | None = None,
    workers: int = 8,
    chunk_mult: int = 2,
    prefetch: int = 2,
    device: str | None = None,
    small_hole_thr: int = 50000,
    buffer: int = 10,
    fraction: float = 0.001,
    nproc: int = 20,
    interior_max_area: int = 100000,
    hematoxylin_channel: bool = False,
    return_raw_gdf: bool = False,
) -> InferenceResult:
    """Inference using SegFormer semantic segmentation model on a WSI.

    Args:
        model (str | torch.nn.Module): Path to the model checkpoint or
            a pre-loaded model.
        wsi_fp (str): File path to the WSI.
        label_ranks (list[int], optional): List of int labels (as
            outputed by the model) ordered by rank with index 0 being
            the lowest rank. If None, The labels will be ranked by
            their int value.
        batch_size (int, optional): Batch size for inference. Defaults
            to 64.
        tile_size (int, optional): Tile size for inference. Defaults to
            512.
        mag (float, optional): Magnification to grab the tiles at for
            inference. This is calculated from the mm_px from the WSI
            metadata, converted to the standard mm_px to magnification
            ratio at Emory. Defaults to 10.0.
        mm_px (float, optional): Micrometers per pixel to grab the tiles
            at for inference. If not provided, will be inferred from the
            mag parameter, if that isn't provided it will use the scan
            resolution of the WSI. Defaults to None.
        workers (int, optional): Number of workers for inference.
            Defaults to 8.
        chunk_mult (int, optional): Chunk multiplier for inference.
            Defaults to 2.
        prefetch (int, optional): Number of prefetch for inference.
            Defaults to 2.
        device (str, optional): Device for inference. Default is None,
            will use "gpu" if available, otherwise "cpu".
        small_hole_thr (int, optional): Threshold in area to identify
            small objects. Defaults to 50000.
        buffer (int, optional): Buffer to add to polygons before
            dissolving. Defaults to 10.
        fraction (float, optional): Fraction of the maximum dimension
            to use for RDP. Defaults to 0.001.
        nproc (int, optional): Number of processes to use for parallel
            RDP. Defaults to 20.
        interior_max_area (int, optional): Maximum area of a hole to fill.
            Used when filling gaps created by RDP. Defaults to 100000.
        hematoxylin_channel (bool, optional): Whether to use the
            hematoxylin channel when predicting the segmentation mask.
            Defaults to False.
        return_raw_gdf (bool, optional): Whether to return the raw
            inference output as a geopandas dataframe. Defaults to False.

    Returns:
        SegFormerSSInferenceResult: Result object containing the inference output.

    """
    # Get the tile source.
    ts = large_image_source_openslide.open(wsi_fp)

    ts_metadata = ts.getMetadata()

    # Calculate the mm per pixel to use.
    if mm_px is None and mag is None:
        print(
            "Using scan resolution, please note we use mm_x and assume mm_y is"
            "the same."
        )
        mm_px = ts_metadata["mm_x"]

        # Calculate what the standard magnification.
        mag = STANDARD_MAG * STANDARD_MM_PER_PX / mm_px
    elif mm_px is None:
        # Calculate the mm per pixel.
        mm_px = STANDARD_MAG * STANDARD_MM_PER_PX / mag
    else:
        # Calculate the magnification.
        mag = STANDARD_MAG * STANDARD_MM_PER_PX / mm_px

    # Scale factor, multiply to go from scan magnification to desired mag.
    sf_x = ts_metadata["mm_x"] / mm_px
    sf_y = ts_metadata["mm_y"] / mm_px

    # Create eager iterator.
    iterator = ts.eagerIterator(
        scale={"mm_x": mm_px, "mm_y": mm_px},
        tile_size={"width": tile_size, "height": tile_size},
        chunk_mult=chunk_mult,
        batch=batch_size,
        prefetch=prefetch,
        workers=workers,
    )

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"

    print(f"Using device: {device}")
    device = torch.device(device)

    # Load the model.
    if isinstance(model, str):
        model = SegformerForSemanticSegmentation.from_pretrained(
            model, local_files_only=True, device_map=device
        )

    model.eval()

    if label_ranks is None:
        id2label = model.config.id2label
        max_label = max([int(v) for v in id2label.keys()])
        label_ranks = list(range(max_label + 1))

    # Iterate through batches.
    batch_n = 0

    # Image processor for images.
    processor = SegformerImageProcessor()

    # Track all predicted polygons.
    wsi_polygons = []

    results = InferenceResult()

    start_time = perf_counter()

    for batch in iterator:
        tiles = batch["tile"].view()  # images are in BCHW format

        if hematoxylin_channel:
            # Deconvolve to get images of the hematoxylin channel.
            img_list = []

            for img in tiles:
                img = (
                    htk.preprocessing.color_deconvolution.color_deconvolution(
                        img, W
                    ).Stains[:, :, 0]
                )
                img = np.stack([img, img, img], axis=-1)
                img_list.append(img)

            tiles = img_list

        # Convert the numpy arrays to PIL images.
        imgs = [Image.fromarray(img) for img in tiles]

        # Pass the images through the processor.
        inputs = processor(imgs, return_tensors="pt")
        inputs = inputs.to(model.device)

        # Predict on the batch.
        with torch.no_grad():
            output = model(inputs["pixel_values"])
            logits = output.logits

            # Get the logits out, resizing them to the original tile size.
            logits = torch.nn.functional.interpolate(
                logits,
                size=tile_size,
                mode="bilinear",
            )

            # Get predicted class labels for each pixel.
            masks = torch.argmax(logits, dim=1).detach().cpu().numpy()

        # Top left corner of each tile, at scan magnification.
        tile_x_coords = batch["gx"]
        tile_y_coords = batch["gy"]

        for i, mask in enumerate(masks):
            x, y = tile_x_coords[i], tile_y_coords[i]

            # Convert top left point to the desired magnification.
            x_scaled = int(x * sf_x)
            y_scaled = int(y * sf_y)

            polygon_and_labels = label_mask_to_polygons(
                mask,
                x_offset=x_scaled,
                y_offset=y_scaled,
            )

            for polygon_and_label in polygon_and_labels:
                polygon, label = polygon_and_label
                label = int(label)

                # Do something with the polygon and label.
                wsi_polygons.append([polygon, label])

        batch_n += 1
        print(f"\r    Processed batch {batch_n}.    ", end="")
    print()

    results.add_time("inference", perf_counter() - start_time)

    # Convert polygons and labels to a GeoDataFrame.
    raw_gdf = gpd.GeoDataFrame(wsi_polygons, columns=["geometry", "label"])
    gdf = raw_gdf.copy()
    if return_raw_gdf:
        # Scale the raw gdf.
        raw_gdf["geometry"] = raw_gdf["geometry"].apply(
            lambda geom: scale(
                geom, xfact=1 / sf_x, yfact=1 / sf_y, origin=(0, 0)
            )
        )

    # Add a small buffer to the polygons to make polygons from adjacent tiles
    # touch, this allows merging adjacent tile polygons when dissolving.
    start_time = perf_counter()
    gdf["geometry"] = gdf["geometry"].buffer(buffer)
    results.add_time("buffer", perf_counter() - start_time)

    start_time = perf_counter()
    gdf = gdf.dissolve(by="label", as_index=False)
    gdf = gdf.explode(index_parts=False).reset_index(drop=True)
    results.add_time("dissolve", perf_counter() - start_time)

    # Scale the geometries.
    gdf["geometry"] = gdf["geometry"].apply(
        lambda geom: scale(geom, xfact=1 / sf_x, yfact=1 / sf_y, origin=(0, 0))
    )

    cleanup_pipe = SegFormerSSInferenceCleanup(
        gdf,
        label_ranks,
        small_hole_thr=small_hole_thr,
        fraction=fraction,
        nproc=nproc,
        interior_max_area=interior_max_area,
    )

    gdf = cleanup_pipe.cleanup()

    for section_name, time in cleanup_pipe.time.items():
        results.add_time(section_name, time)

    results.add_field("gdf", gdf)
    results.add_field("mag", mag)
    results.add_field("mm_px", mm_px)

    if return_raw_gdf:
        results.add_field("raw_gdf", raw_gdf)

    return results


class SegFormerSSInferenceCleanup:
    def __init__(
        self,
        gdf: gpd.GeoDataFrame,
        label_ranks: list[int],
        small_hole_thr: int = 50000,
        buffer: int = 1,
        fraction: float = 0.001,
        nproc: int = 20,
        interior_max_area: int = 100000,
    ):
        """Initiate the class for cleaning up the inference output.

        Args:
            gdf (geopandas.GeoDataFrame): Input inference output.
            label_ranks (list[int]): List of labels, ordered by rank
                with index 0 being the lowest rank.
            small_hole_thr (int, optional): Threshold in area to
                identify small objects. Defaults to 50000.
            buffer (int, optional): Buffer to add to polygons before
                dissolving. Defaults to 1.
            fraction (float, optional): Fraction of the maximum
                dimension to use for RDP. Defaults to 0.001.
            nproc (int, optional): Number of processes to use for
                parallel RDP. Defaults to 20.
            interior_max_area (int, optional): Maximum area of a hole
                to fill. Used when filling gaps created by RDP.
                Defaults to 100000.

        """
        for i, r in gdf.iterrows():
            label = r["label"]

            if label not in label_ranks:
                raise ValueError(f"Label {label} not in label_ranks.")
            gdf.loc[i, "rank"] = label_ranks.index(r["label"])

        self.__version__ = "1.0.2"
        self.input_gdf = gdf
        self.small_hole_thr = small_hole_thr
        self.output_gdf = None
        self.fraction = fraction
        self.nproc = nproc
        self.interior_max_area = interior_max_area
        self.time = {}

    def _make_gpd_valid(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        # Make the geometries valid in the gdf, keep only polygons.
        gdf["geometry"] = gdf["geometry"].apply(make_valid)
        gdf = gdf.explode(index_parts=False)
        gdf = gdf[
            (gdf["geometry"].geom_type == "Polygon")
            & (gdf["geometry"].is_valid)
        ]

        return gdf.reset_index(drop=True)

    def _remove_intersections(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        # Remove intersections between polygons.
        gdf = gdf.reset_index(drop=True)
        n = len(gdf)

        if n in (0, 1):
            return gdf

        # Loop until the second to last row.
        for i in tqdm(
            range(n - 1), total=n - 1, desc="Removing intersections"
        ):
            r1 = gdf.iloc[i]

            # Subtract the r1 geometry from all others.
            for j in range(i + 1, n):
                r2 = gdf.iloc[j]

                # Subtract r1 from r2.
                geom = r2["geometry"].difference(r1["geometry"])

                gdf.loc[j, "geometry"] = geom

        return self._make_gpd_valid(gdf)

    def _remove_small_holes(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        gdf = gdf.reset_index(drop=True)

        n = len(gdf)

        for i in tqdm(range(n), total=n, desc="Removing small holes"):
            geom = gdf.iloc[i]["geometry"]

            exterior = geom.exterior
            interiors = geom.interiors

            new_interiors = []
            for interior in interiors:
                if Polygon(interior).area > self.small_hole_thr:
                    new_interiors.append(interior)

            geom = Polygon(exterior, new_interiors)

            gdf.loc[i, "geometry"] = geom

        return gdf

    def _remove_small_contained_polygons(
        self, gdf: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        n = len(gdf)

        i_removed = []

        for i in tqdm(
            range(n), total=n, desc="Removing small contained polygons"
        ):
            exterior = gdf.iloc[i]["geometry"].exterior

            geom = Polygon(exterior)

            if geom.area < self.small_hole_thr:
                # Check if this is contained in another polygon.
                contained = gdf[
                    (gdf["geometry"].contains(geom)) & (gdf.index != i)
                ]

                if len(contained):
                    i_removed.append(i)

        gdf = gdf.drop(i_removed)

        return gdf

    def _remove_small_polygons_not_contained(
        self, gdf: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        gdf = gdf.reset_index(drop=True)
        n = len(gdf)

        i_removed = []

        for i in tqdm(
            range(n), total=n, desc="Removing small polygons not contained"
        ):
            geom = gdf.iloc[i]["geometry"]

            if geom.geom_type == "MultiPolygon":
                area = geom.area
            else:
                area = Polygon(geom.exterior).area

            if area < self.small_hole_thr:
                # Check for any touching polygons.
                touching = gdf[
                    (~gdf.index.isin(i_removed + [i]))
                    & (gdf["geometry"].touches(geom))
                ].copy()
                if len(touching):
                    touching["intersection_length"] = (
                        touching["geometry"].intersection(geom).length
                    )

                    touching = touching.sort_values(
                        by="intersection_length", ascending=False
                    )

                    r = touching.iloc[0]

                    touching_geom = r["geometry"]

                    # Merge the polygons.
                    geom = geom.union(touching_geom)

                    gdf.loc[r.name, "geometry"] = geom
                    i_removed.append(i)
                else:
                    # Remove this polygon.
                    i_removed.append(i)

        gdf = gdf.drop(i_removed)
        gdf = self._make_gpd_valid(gdf)
        return gdf

    def _rdp_polygon(self, geom, idx, fraction):
        geom = rdp_by_fraction_of_max_dimension(geom, fraction=fraction)
        return geom, idx

    def _fill_rdp_gaps(
        self, gdf: gpd.GeoDataFrame, interior_max_area: int = 100000
    ) -> gpd.GeoDataFrame:
        # Fill the gaps between polygons created by RDP.
        gdf["area"] = gdf["geometry"].area

        union_geom = gdf["geometry"].union_all()

        interiors = []

        if union_geom.geom_type == "MultiPolygon":
            # Collect all the holes / interiors.
            for geom in union_geom.geoms:
                for interior in geom.interiors:
                    interior = Polygon(interior)

                    if interior.area < interior_max_area:
                        interiors.append(interior)
        elif union_geom.geom_type == "Polygon":
            for interior in union_geom.interiors:
                interior = Polygon(interior)

                if interior.area < interior_max_area:
                    interiors.append(interior)
        else:
            raise ValueError(
                f"Unexpected geometry type: {union_geom.geom_type}"
            )

        # Loop through each hole that was small.
        for interior in tqdm(interiors, desc="Filling holes between polygons"):
            # Check polygons that are touching this interior.
            touching = gdf[gdf["geometry"].distance(interior) == 0]

            unique_ranks = touching["rank"].unique()

            if len(unique_ranks) > 1:
                # Sort by rank then area.
                touching = touching.sort_values(
                    by=["rank", "area"], ascending=False
                )

                # Get the first row.
                r = touching.iloc[0]

                # Merge the hole with the polygon.
                geom = unary_union([r["geometry"], interior])

                if geom.geom_type == "MultiPolygon":
                    # Buff the interior a bit.
                    interior_buffed = interior.buffer(1)
                    geom = unary_union([interior_buffed, geom])

                    if geom.geom_type == "MultiPolygon":
                        print(
                            "MultiPolygon after buffering and union, discarding hole."
                        )
                        continue

                gdf.loc[r.name, "geometry"] = geom
                gdf.loc[r.name, "area"] = geom.area

        return gdf

    def cleanup(self):
        """Pipeline for cleaning up the inference output."""
        print("Running inference cleanup:\n")
        time = self.time
        gdf = self.input_gdf.copy()
        gdf = self._make_gpd_valid(gdf)

        print("[1/7] Removing intersections...")
        start_time = perf_counter()
        gdf = make_multi_polygons(gdf, "label")
        gdf = self._remove_intersections(gdf)
        time["remove-intersections"] = perf_counter() - start_time

        print("[2/7] Removing small holes...")
        gdf = gdf.explode(index_parts=False).reset_index(drop=True)
        gdf = self._make_gpd_valid(gdf)
        start_time = perf_counter()
        gdf = self._remove_small_holes(gdf)
        time["remove-small-holes"] = perf_counter() - start_time

        print("[3/7] Removing small polygons contained in other polygons...")
        start_time = perf_counter()
        gdf = self._remove_small_contained_polygons(gdf)
        time["remove-small-contained-polygons"] = perf_counter() - start_time

        print("[4/7] Removing small polygons not contained...")
        start_time = perf_counter()
        gdf = self._remove_small_polygons_not_contained(gdf)
        time["remove-small-polygons-not-contained"] = (
            perf_counter() - start_time
        )

        # Parallel RDP.
        print("[5/7] Reducing points in polygons via RDP...")
        start_time = perf_counter()

        with Pool(processes=self.nproc) as pool:
            jobs = [
                pool.apply_async(
                    func=self._rdp_polygon,
                    args=(r["geometry"], i, self.fraction),
                )
                for i, r in gdf.iterrows()
            ]

            n = len(gdf)
            completed = 0

            # Process jobs as they become ready
            with tqdm(total=n, desc="Reducing points in polygons") as pbar:
                while completed < n:
                    for job in jobs:
                        if job.ready():
                            geom, idx = job.get()
                            gdf.loc[idx, "geometry"] = geom
                            completed += 1
                            pbar.update(1)
                            # Remove completed job from list to avoid checking it again
                            jobs.remove(job)
                            break
                    else:
                        # Small sleep to avoid busy waiting
                        sleep(0.01)

        time["rdp"] = perf_counter() - start_time

        gdf = self._make_gpd_valid(gdf)

        print("[6/7] Removing intersections again...")
        start_time = perf_counter()
        gdf = make_multi_polygons(gdf, "label")
        gdf = self._remove_intersections(gdf)
        time["remove-intersections-2"] = perf_counter() - start_time

        gdf = gdf.explode(index_parts=False).reset_index(drop=True)
        gdf = self._make_gpd_valid(gdf)

        print("[7/7] Filling RDP gaps...")
        start_time = perf_counter()
        gdf = self._fill_rdp_gaps(gdf, self.interior_max_area)
        time["fill-rdp-gaps"] = perf_counter() - start_time
        gdf = self._make_gpd_valid(gdf)

        self.output_gdf = gdf
        self.time = time

        total_time = sum(time.values())
        print(f"\nTotal time: {total_time:.2f} seconds")
        return gdf



import yaml, shutil
from pathlib import Path
from time import perf_counter
from ultralytics import YOLO

from ...utils import convert_to_json_serializable


def train_yolo(
    project_dir: str,
    model_name: str,
    yaml_path: str,
    validate_with_train: bool = True,
    validate_with_val: bool = True,
    pretrained_model: str = "yolo11m",
    epochs: int = 100,
    patience: int = 10,
    batch_size: int = 16,
    imgsz: int = 640,
    device: int | str | list | None = None,
    workers: int = 8,
    lr0: float = 0.01,
    train_kwargs: dict = None,
    agnostic_nms: bool = True,
    conf: float = 0.001,
    iou: float = 0.7,
):
    """Function for training a YOLO model with the Ultralytics API.

    * Some of the argument descriptions are taken from the Ultralytics
    documentation.

    Args:
        project_dir (str): Parent directory the model directory will be
            saved in.
        model_name (str): The name of the model to train.
        yaml_path (str): Path do dataset yaml file that specifies images
            that will be used in train and val, and classes.
        validate_with_train (bool, optional): True to validate on the
            train dataset after training. Defaults to True.
        validate_with_val (bool, optional): True to validate on the
            val dataset after training. Defaults to True.
        pretrained_model (str, optional): The pretrained model to use.
            Local or from Ultralytics model hub. Defaults to "yolo11m".
        epochs (int, optional): The number of epochs to train for.
            Defaults to 100.
        patience (int, optional): The number of epochs to wait before
            early stopping. Defaults to 10.
        batch_size (int, optional): The batch size to use for training.
            Defaults to 16.
        imgsz (int, optional): The size of the images to use for
            training. Defaults to 640.
        device (int | str | list | None, optional): The device to use
            for training. Defaults to None.
        workers (int, optional): The number of workers to use for
            training. Defaults to 8.
        lr0 (float, optional): Initial training learning rate. Defaults
            to 0.01.
        augmentation_kwargs (dict, optional):
            flip_lr (float, optional): The probability of flipping the
                image horizontally. Defaults to 0.5.
            flip_ud (float, optional): The probability of flipping the
                image vertically. Defaults to 0.5.
            degrees (float, optional): The number of degrees to rotate
                the image. Defaults to 0.2.
            shear (float, optional): The amount of shear to apply to the
                image. Defaults to 0.2.
            mosaic (float, optional): The probability of applying a
                mosaic augmentation to the image. Defaults to 0.1.
            hsv_s (float, optional): The saturation of the image.
                Defaults to 0.6.
            hsv_v (float, optional): The intensity of the image.
                Defaults to 0.6.
            scale (float, optional): The amount of zoom to apply to the
                image. Defaults to 0.2.
        agnostic_nms (bool, optional): Enables class-agnostic
            Non-Maximum Suppression, which merges overlapping boxes
            regardless of their predicted class. Useful for
            instance-focused applications. Defaults to True.
        conf (float, optional): Sets the minimum confidence threshold
            for detections. Lower values increase recall but may
            introduce more false positives. Used during validation to
            compute precision-recall curves. Defaults to 0.001.
        iou (float, optional): Sets the Intersection Over Union
            threshold for Non-Maximum Suppression. Controls duplicate
            detection elimination. Defaults to 0.7.

    Returns:
        model: The trained model.
        results: The results of the training.
        validation_results (dict): The results of the validation.
        train_time (float): Time to train model in seconds.

    """
    project_dir_path = Path(project_dir)

    if not project_dir_path.is_absolute():
        raise ValueError(f"Project directory {project_dir} is not absolute.")

    # Load the model.
    model = YOLO(pretrained_model)

    if train_kwargs is None:
        train_kwargs = {}

    train_augmentation = {
        "fliplr": train_kwargs.get("hflip", 0.5),
        "flipud": train_kwargs.get("vflip", 0.5),
        "degrees": train_kwargs.get("degrees", 0.2),
        "shear": train_kwargs.get("shear", 0.2),
        "mosaic": train_kwargs.get("mosaic", 0.1),
        "hsv_s": train_kwargs.get("saturation", 0.6),
        "hsv_v": train_kwargs.get("intensity", 0.6),
        "scale": max(
            train_kwargs.get("zoom_in", 0.2), train_kwargs.get("zoom_out", 0.2)
        ),
    }

    # Create the project directory.
    project_dir = Path(project_dir)
    project_dir.mkdir(parents=True, exist_ok=True)

    model_dir = project_dir / model_name

    if model_dir.is_dir():
        raise FileExistsError(
            f"Model {model_name} already exists in {project_dir}."
        )

    # Read the YAML file.
    with open(yaml_path, "r") as f:
        yaml_data = yaml.safe_load(f)

    nc = yaml_data["nc"]
    names = yaml_data["names"]

    # Train the model.
    start_time = perf_counter()
    results = model.train(
        data=str(yaml_path),
        epochs=epochs,
        batch=batch_size,
        project=str(project_dir),
        name=model_name,
        patience=patience,
        imgsz=imgsz,
        device=device,
        workers=workers,
        lr0=lr0,
        agnostic_nms=agnostic_nms,
        **train_augmentation,
    )
    train_time = perf_counter() - start_time

    validation_results = {}
    runs_dir = project_dir / "runs"

    if device is not None and isinstance(device, list):
        device_str = [f"cuda:{d}" for d in device]
        device_str = ",".join(device_str)
    else:
        device_str = device

    if (validate_with_val or validate_with_train) and results is None:
        # This is for multi-processing.
        print(
            "Warning: model.train() returned None (common with DDP). Reloading results from saved model..."
        )

        best_model_path = model_dir / "weights" / "best.pt"

        if best_model_path.exists():
            # Reload the model and get results by running validation
            model = YOLO(str(best_model_path))
            results = model.val(
                data=str(yaml_path),
                split="val",
                batch=batch_size,
                device=device_str,
                workers=workers,
                agnostic_nms=agnostic_nms,
                conf=conf,
                iou=iou,
            )
            print("Successfully reloaded results from saved model.")
        else:
            raise FileNotFoundError(
                f"Best model path {best_model_path} does not exist."
            )

    if validate_with_train:
        runs_dir.mkdir(parents=True, exist_ok=True)
        validation_results["train"] = {}

        # Run validation on each label.
        for i in range(nc):
            print(f"Validating on train for class {names[i]}...")
            metrics = model.val(
                split="train",
                classes=[i],
                project=str(runs_dir),
                batch=batch_size,
                device=device_str,
                workers=workers,
                agnostic_nms=agnostic_nms,
                conf=conf,
                iou=iou,
            )

            validation_results["train"][names[i]] = (
                convert_to_json_serializable(metrics.results_dict)
            )

    if validate_with_val:
        runs_dir.mkdir(parents=True, exist_ok=True)
        validation_results["val"] = {}

        # Run validation on each label.
        for i in range(nc):
            print(f"Validating on val for class {names[i]}...")
            metrics = model.val(
                split="val",
                classes=[i],
                project=str(runs_dir),
                batch=batch_size,
                device=device_str,
                workers=workers,
                agnostic_nms=agnostic_nms,
                conf=conf,
                iou=iou,
            )

            validation_results["val"][names[i]] = convert_to_json_serializable(
                metrics.results_dict
            )

    # Remove the runs dir.
    if runs_dir.is_dir():
        shutil.rmtree(runs_dir)

    return model, results, validation_results, train_time



from dataclasses import dataclass, field
from typing import Any


@dataclass
class InferenceResult:
    """Class for storing inference results."""

    time: dict[str, Any] = field(
        default_factory=lambda: {"total": 0.0, "sections": {}},
        kw_only=True,  # Makes this field keyword-only, so it doesn't conflict with positional args in subclasses
    )

    fields: dict[str, Any] = field(default_factory=dict, kw_only=True)

    def add_time(self, section_name: str, value: float) -> None:
        """Add a time measurement to the sections dictionary and update total time.

        Args:
            section_name: Name of the section (e.g., "model_loading", "inference")
            value: Time in seconds for this section
        """
        self.time["sections"][section_name] = value
        self.time["total"] = sum(self.time["sections"].values())

    def get_total_time(self) -> float:
        """Get the total time from all sections."""
        return self.time["total"]

    def get_sections(self) -> dict[str, float]:
        """Get the sections time dictionary."""
        return self.time["sections"]

    def add_field(self, field_name: str, value: Any) -> None:
        """Add a field to the result."""
        self.fields[field_name] = value

    def __getattr__(self, name: str) -> Any:
        """Allow accessing fields as attributes."""
        if name in self.fields:
            return self.fields[name]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def __setattr__(self, name: str, value: Any) -> None:
        """Allow setting fields as attributes if they're not dataclass fields."""
        # Check if it's a dataclass field
        if (
            hasattr(self.__class__, "__dataclass_fields__")
            and name in self.__class__.__dataclass_fields__
        ):
            super().__setattr__(name, value)
        else:
            # It's a dynamic field
            if not hasattr(self, "fields"):
                super().__setattr__("fields", {})
            self.fields[name] = value

    def to_dict(self) -> dict[str, Any]:
        """Convert the result to a dictionary."""
        result = {
            "time": self.time,  # Keep the full nested time structure
        }
        # Add all dynamic fields
        result.update(self.fields)
        return result



"""Tiling function for use in YOLO object detection from Ultrayltics"""

import large_image, large_image_source_openslide
import geopandas as gpd
import cv2 as cv
import pandas as pd
from pathlib import Path
from tqdm import tqdm

from shapely import force_2d, Polygon, box
from shapely.affinity import translate, scale

from ...utils import return_mag_and_resolution
from ... import imwrite


def tile_wsi_for_yolo(
    wsi_fp: str,
    geojson_doc: dict,
    roi_labels: str | list[str],
    box_labels: list[str],
    label2id: dict[str, int],
    save_dir: str,
    magnification: int = 20,
    tile_size: int = 1280,
    stride: int | None = 960,
    tile_area_threshold: float = 0.25,
    pad_rgb: tuple[int, int, int] = (114, 114, 114),
    box_area_threshold: float = 0.5,
    allow_small_rotations: bool = True,
    small_rotation_threshold: float = 5,
) -> pd.DataFrame:
    """Tile WSI for YOLO model training in the Ultralytics format.

    Args:
        wsi_fp (str): File path to the WSI.
        geojson_doc (dict): DSA annotation document in GeoJSON format.
        roi_labels (str | list[str]): Label(s) of the ROI annotations.
            For now only rectangular ROIs with no rotation are supported,
            others are ignored.
        box_labels (list[str]): Labels of the object bounding box
            annotations. The integer labels are determined by the order
            of the labels on the list. Note that again only rectangular
            boxes with no rotation are supported, others are ignored.
        label2id (dict[str, int]): Mapping of label names to integer
            indices. All box labels must be in the label2id dictionary.
        save_dir (str): Directory to save the tiles and labels to.
        magnification (int, optional): Magnification to get tiles at.
            Defaults to 20.
        tile_size (int, optional): Size of the tiles to use. Defaults to
            1280.
        stride (int | None, optional): Stride to use for the tiles. If
            None, the stride is set to the tile size. Defaults to 960.
        tile_area_threshold (float, optional): Fraction of tile that
            must be in the ROI to be included. If the fraction is less
            than this value than it is ignored. Defaults to 0.25.
        pad_rgb (tuple[int, int, int], optional): RGB values to pad the
            tile with if it is smaller than the tile size. Defaults to
            (114, 114, 114).
        box_area_threshold (float, optional): Fraction of box that must
            be in the tile to be included. If the fraction is less than
            this value than it is ignored. Defaults to 0.5.
        allow_small_rotations (bool, optional): Whether to allow small
            rotations of the boxes. Rotations less than 5 degrees are
            ignored, and the smallest bounding box is used instead on
            the tile. Defaults to True.
        small_rotation_threshold (float, optional): The threshold for
            small rotations. Defaults to 5.

    Returns:
        pandas.DataFrame: DataFrame with the tile metadata.

    """
    assert len(box_labels) == len(
        set(box_labels)
    ), "Box labels must be unique."

    for label in box_labels:
        assert label in label2id, f"Label {label} not in label2id."

    if isinstance(roi_labels, str):
        roi_labels = [roi_labels]

    # Make the labels strings.
    roi_labels = [str(label) for label in roi_labels]
    box_labels = [str(label) for label in box_labels]

    save_dir_path = Path(save_dir)
    img_dir = save_dir_path / "images"
    img_dir.mkdir(parents=True, exist_ok=True)
    label_dir = save_dir_path / "labels"
    label_dir.mkdir(parents=True, exist_ok=True)

    gdf = gpd.GeoDataFrame.from_features(geojson_doc["features"])
    gdf["geometry"] = gdf["geometry"].apply(force_2d)  # remove z-coordinate

    # Convert the label column, which is a dict to use the value as cell value.
    gdf["label"] = gdf["label"].apply(lambda x: x["value"])

    # Filter out elements that are not part of ROI or box elements.
    gdf = gdf[gdf["label"].isin(roi_labels + box_labels)]

    # Split into rectangle and polygon elements.
    rect_gdf = gdf[gdf["type"] == "rectangle"].reset_index(drop=True)
    poly_gdf = gdf[gdf["type"] == "polyline"].reset_index(drop=True)

    # Modify rotated rectangles if allowed.
    if len(rect_gdf):
        if allow_small_rotations:
            # Remove rows with rotations less than 5 degrees.
            rect_gdf = rect_gdf[
                rect_gdf["rotation"] <= small_rotation_threshold
            ]

            # Take only the smallest bounding box for each box.
            rect_gdf["geometry"] = rect_gdf["geometry"].apply(
                lambda geom: Polygon(
                    [
                        (geom.bounds[0], geom.bounds[1]),  # (minx, miny)
                        (geom.bounds[2], geom.bounds[1]),  # (maxx, miny)
                        (geom.bounds[2], geom.bounds[3]),  # (maxx, maxy)
                        (geom.bounds[0], geom.bounds[3]),  # (minx, maxy)
                        (geom.bounds[0], geom.bounds[1]),  # close the polygon
                    ]
                )
            )
        else:
            rect_gdf = rect_gdf[(rect_gdf["rotation"] == 0)]
    else:
        rect_gdf = gpd.GeoDataFrame()

    if len(poly_gdf):
        # Handle polylines.
        for i, r in poly_gdf.iterrows():
            poly = r["geometry"]

            minx, miny, maxx, maxy = poly.bounds
            rectangle = box(minx, miny, maxx, maxy)
            gdf.loc[i, "geometry"] = rectangle
    else:
        poly_gdf = gpd.GeoDataFrame()

    # Concatenate the rectangle and polygon dataframes.
    gdf = pd.concat([rect_gdf, poly_gdf], ignore_index=True)

    metadata = []

    tile_thr = tile_size * tile_size * tile_area_threshold

    if len(gdf):
        roi_gdf = gdf[gdf["label"].isin(roi_labels)].reset_index(drop=True)
        box_gdf = gdf[gdf["label"].isin(box_labels)].reset_index(drop=True)

        # Get the large image tile source.
        ts = large_image_source_openslide.open(wsi_fp)
        wsi_name = Path(wsi_fp).stem

        if stride is None:
            stride = tile_size

        magnification, mm_px = return_mag_and_resolution(mag=magnification)

        # Calculate the scale factor on each axis.
        # Calculating a multirplicative factor for going from scan mag to desired mag.
        # With mm_px, higher values are more zoomed in, so we reverse the division.
        ts_metadata = ts.getMetadata()
        sf_x = ts_metadata["mm_x"] / mm_px
        sf_y = ts_metadata["mm_y"] / mm_px

        n_rois = len(roi_gdf)

        # Loop through each ROI.
        for i, roi_row in roi_gdf.iterrows():
            print(f"Processing ROI {i+1} of {n_rois}...")
            # Get the bounds of the ROI.
            roi_x1, roi_y1, roi_x2, roi_y2 = roi_row["geometry"].bounds
            roi_x1, roi_y1 = int(roi_x1), int(roi_y1)
            roi_x2, roi_y2 = int(roi_x2), int(roi_y2)

            roi_key = f"x{roi_x1}y{roi_y1}x{roi_x2}y{roi_y2}"

            # Make a copy of the label boxes.
            box_gdf_copy = box_gdf.copy()

            # Shift the boxes so zero zero is the top left of the ROI.
            box_gdf_copy["geometry"] = box_gdf_copy["geometry"].apply(
                lambda x: translate(x, xoff=-roi_x1, yoff=-roi_y1)
            )

            # Scale the boxes so they are at the desired magnification.
            box_gdf_copy["geometry"] = box_gdf_copy["geometry"].apply(
                lambda x: scale(x, xfact=sf_x, yfact=sf_y, origin=(0, 0))
            )

            # Get the ROI at the specified resolution.
            roi_img = ts.getRegion(
                region={
                    "left": roi_x1,
                    "top": roi_y1,
                    "right": roi_x2,
                    "bottom": roi_y2,
                },
                format=large_image.constants.TILE_FORMAT_NUMPY,
                scale={"mm_x": mm_px, "mm_y": mm_px},
            )[0][:, :, :3].copy()

            # Calculate the top left coordinates of each tile in this ROI.
            xys = []
            roi_h, roi_w = roi_img.shape[:2]

            for x in range(0, roi_w, stride):
                for y in range(0, roi_h, stride):
                    xys.append((x, y))

            # Loop through each tile.
            for xy in tqdm(xys, desc="Processing tiles"):
                x, y = xy

                # Get the tile image.
                tile_img = roi_img[y : y + tile_size, x : x + tile_size]

                # Shift all the boxes by the top left corner of this tile.
                box_gdf_shifted = box_gdf_copy.copy()
                box_gdf_shifted["geometry"] = box_gdf_shifted[
                    "geometry"
                ].apply(lambda geom: translate(geom, xoff=-x, yoff=-y))

                # Add the area of the box to tile.
                box_gdf_shifted["area"] = box_gdf_shifted["geometry"].area

                # Create a polygon for the tile
                tile_h, tile_w = tile_img.shape[:2]

                # Determine if the tile is too small.
                tile_area = tile_h * tile_w
                if tile_area < tile_thr:
                    # Skip the tile.
                    continue

                tile_box = Polygon(
                    [
                        (0, 0),
                        (tile_w, 0),
                        (tile_w, tile_h),
                        (0, tile_h),
                        (0, 0),
                    ]
                )

                # Calculate the intersection.
                box_gdf_shifted["geometry"] = box_gdf_shifted[
                    "geometry"
                ].intersection(tile_box)

                # Remove the empty geometries.
                box_gdf_shifted = box_gdf_shifted[
                    box_gdf_shifted["geometry"].area > 0
                ].reset_index(drop=True)

                # Calculate the new area.
                box_gdf_shifted["clipped_area"] = box_gdf_shifted[
                    "geometry"
                ].area

                # Remove boxes that are less than the threshold
                box_gdf_shifted["frac_in_tile"] = (
                    box_gdf_shifted["clipped_area"] / box_gdf_shifted["area"]
                )
                box_gdf_shifted = box_gdf_shifted[
                    box_gdf_shifted["frac_in_tile"] >= box_area_threshold
                ].reset_index(drop=True)

                if tile_h != tile_size or tile_w != tile_size:
                    # Pad the edge of the tile.
                    tile_img = cv.copyMakeBorder(
                        tile_img,
                        0,
                        tile_size - tile_h,
                        0,
                        tile_size - tile_w,
                        cv.BORDER_CONSTANT,
                        value=pad_rgb,
                    )

                tile_key = f"{wsi_name}_{roi_key}_x{x}y{y}x{x+tile_size}y{y+tile_size}"
                img_fp = img_dir / f"{tile_key}.png"
                imwrite(img_fp, tile_img)

                # Write the labels to file.
                labels = ""

                for _, r in box_gdf_shifted.iterrows():
                    # Get the width and height of the object.
                    bx1, by1, bx2, by2 = r.geometry.bounds

                    xc = (bx1 + bx2) / 2
                    yc = (by1 + by2) / 2
                    bh = by2 - by1
                    bw = bx2 - bx1

                    # normalize
                    xc /= tile_size
                    yc /= tile_size
                    bw /= tile_size
                    bh /= tile_size

                    labels += f"{box_labels.index(r['label'])} {xc:.4f} {yc:.4f} {bw:.4f} {bh:.4f}\n"

                label_fp = label_dir / f"{tile_key}.txt"

                if labels:
                    with open(label_fp, "w") as f:
                        f.write(labels.strip())

                metadata.append(
                    [
                        str(img_fp),
                        str(label_fp),
                        roi_key,
                        x,
                        y,
                        x + tile_size,
                        y + tile_size,
                        magnification,
                        mm_px,
                    ]
                )
    else:
        metadata = []

    return pd.DataFrame(
        metadata,
        columns=[
            "img_fp",
            "label_fp",
            "roi_key",
            "x1",
            "y1",
            "x2",
            "y2",
            "mag",
            "mm_px",
        ],
    )



# Function to save image to disk.
import cv2 as cv
import numpy as np


def imwrite(fp: str, img: np.ndarray, grayscale: bool = False) -> None:
    """Save an image to disk.

    Args:
        img (np.ndarray): The image to save.
        fp (str): The file path to save the image.
        grayscale (bool, optional): Whether to save the image as grayscale.
            Defaults to False.

    Raises:
        ValueError: The image must be grayscale, RGB, or RGBA.

    """
    if grayscale:
        if len(img.shape) == 2:
            _ = cv.imwrite(fp, img)
        else:
            _ = cv.imwrite(fp, cv.cvtColor(img, cv.COLOR_RGB2GRAY))
    else:
        shape = img.shape

        if len(shape) == 2:
            _ = cv.imwrite(fp)
        elif len(shape) == 3:
            if shape[2] == 3:
                _ = cv.imwrite(fp, cv.cvtColor(img, cv.COLOR_RGB2BGR))
            elif shape[2] == 4:
                _ = cv.imwrite(fp, cv.cvtColor(img, cv.COLOR_RGBA2BGRA))
            else:
                raise ValueError("The image must be grayscale, RGB, or RGBA.")
        else:
            raise ValueError("The image must be grayscale, RGB, or RGBA.")
